"""Enlarge kernels.lang column length

Revision ID: 10e39a34eed5
Revises: d582942886ad
Create Date: 2018-10-29 13:52:10.583443

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "10e39a34eed5"
down_revision = "d582942886ad"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "kernels", "lang", existing_type=sa.String(length=64), type_=sa.String(length=512)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "kernels", "lang", existing_type=sa.String(length=512), type_=sa.String(length=64)
    )
    # ### end Alembic commands ###



"""add_environ_to_kernels

Revision ID: c3e74dcf1808
Revises: d52bf5ec9ef3
Create Date: 2017-11-15 11:31:54.083566

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c3e74dcf1808"
down_revision = "d52bf5ec9ef3"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("kernels", sa.Column("environ", sa.ARRAY(sa.String()), nullable=True))


def downgrade():
    op.drop_column("kernels", "environ")



from __future__ import annotations

from typing import Optional, Sequence

import trafaret as t

from ai.backend.common.types import (
    AccessKey,
    AgentId,
    AgentSelectionStrategy,
    ResourceSlot,
    SessionId,
)

from ..models import AgentRow, SessionRow
from .types import AbstractScheduler, KernelInfo


class MOFScheduler(AbstractScheduler):
    """Minimum Occupied slot First Scheduler"""

    config_iv = t.Dict({}).allow_extra("*")

    def pick_session(
        self,
        total_capacity: ResourceSlot,
        pending_sessions: Sequence[SessionRow],
        existing_sessions: Sequence[SessionRow],
    ) -> Optional[SessionId]:
        # Just pick the first pending session.
        return SessionId(pending_sessions[0].id)

    def _assign_agent(
        self,
        agents: Sequence[AgentRow],
        access_key: AccessKey,
        requested_slots: ResourceSlot,
        agent_selection_strategy: AgentSelectionStrategy,
        agent_selection_resource_priority: list[str],
    ) -> Optional[AgentId]:
        # return min occupied slot agent or None
        return next(
            (
                one_agent.id
                for one_agent in (
                    sorted(
                        (
                            agent
                            for agent in agents
                            if ((agent.available_slots - agent.occupied_slots) >= requested_slots)
                        ),
                        key=lambda agent: agent.occupied_slots,
                    )
                )
            ),
            None,
        )

    def assign_agent_for_session(
        self,
        agents: Sequence[AgentRow],
        pending_session: SessionRow,
        agent_selection_strategy: AgentSelectionStrategy,
        agent_selection_resource_priority: list[str],
    ) -> Optional[AgentId]:
        return self._assign_agent(
            agents,
            pending_session.access_key,
            pending_session.requested_slots,
            agent_selection_strategy,
            agent_selection_resource_priority,
        )

    def assign_agent_for_kernel(
        self,
        agents: Sequence[AgentRow],
        pending_kernel: KernelInfo,
        agent_selection_strategy: AgentSelectionStrategy,
        agent_selection_resource_priority: list[str],
    ) -> Optional[AgentId]:
        return self._assign_agent(
            agents,
            pending_kernel.access_key,
            pending_kernel.requested_slots,
            agent_selection_strategy,
            agent_selection_resource_priority,
        )



"""enumerate_vfolder_perms

Revision ID: a1fd4e7b7782
Revises: f9971fbb34d9
Create Date: 2018-09-05 16:51:49.973195

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

from ai.backend.manager.models import VFolderPermission

# revision identifiers, used by Alembic.
revision = "a1fd4e7b7782"
down_revision = "f9971fbb34d9"
branch_labels = None
depends_on = None

# NOTE: VFolderPermission is EnumValueType
vfperm_choices = list(map(str, VFolderPermission))
vfolderpermission = postgresql.ENUM(
    *vfperm_choices,
    name="vfolderpermission",
)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    vfolderpermission.create(op.get_bind())
    op.alter_column(
        "vfolder_invitations",
        column_name="permission",
        type_=sa.Enum(*vfperm_choices, name="vfolderpermission"),
        postgresql_using="permission::vfolderpermission",
    )
    op.alter_column(
        "vfolder_permissions",
        column_name="permission",
        type_=sa.Enum(*vfperm_choices, name="vfolderpermission"),
        postgresql_using="permission::vfolderpermission",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "vfolder_invitations",
        column_name="permission",
        type_=sa.String(length=2),
        postgresql_using="permission::text::vfolderpermission",
    )
    op.alter_column(
        "vfolder_permissions",
        column_name="permission",
        type_=sa.String(length=2),
        postgresql_using="permission::text::vfolderpermission",
    )
    vfolderpermission.drop(op.get_bind())
    # ### end Alembic commands ###



from __future__ import annotations

import asyncio
import hashlib
import itertools
import json
import logging
import uuid
from contextvars import ContextVar
from datetime import datetime, timedelta
from decimal import Decimal
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Final,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import aiotools
import async_timeout
import sqlalchemy as sa
from dateutil.tz import tzutc
from redis.asyncio import Redis
from redis.asyncio.client import Pipeline as RedisPipeline
from sqlalchemy.exc import DBAPIError
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import noload, selectinload

from ai.backend.common import redis_helper
from ai.backend.common.defs import REDIS_LIVE_DB
from ai.backend.common.distributed import GlobalTimer
from ai.backend.common.events import (
    AgentStartedEvent,
    CoalescingOptions,
    DoPrepareEvent,
    DoScaleEvent,
    DoScheduleEvent,
    EventDispatcher,
    EventProducer,
    KernelLifecycleEventReason,
    RouteCreatedEvent,
    SessionCancelledEvent,
    SessionEnqueuedEvent,
    SessionPreparingEvent,
    SessionScheduledEvent,
    SessionTerminatedEvent,
)
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.plugin.hook import PASSED, HookResult
from ai.backend.common.types import (
    AgentId,
    ClusterMode,
    RedisConnectionInfo,
    ResourceSlot,
    RoundRobinState,
    SessionId,
    aobject,
)
from ai.backend.manager.models.session import _build_session_fetch_query
from ai.backend.manager.types import DistributedLockFactory
from ai.backend.plugin.entrypoint import scan_entrypoints

from ..api.exceptions import GenericBadRequest, InstanceNotAvailable, SessionNotFound
from ..defs import SERVICE_MAX_RETRIES, LockID
from ..exceptions import convert_to_status_data
from ..models import (
    AgentRow,
    AgentStatus,
    EndpointLifecycle,
    EndpointRow,
    KernelRow,
    KernelStatus,
    RouteStatus,
    RoutingRow,
    ScalingGroupOpts,
    ScalingGroupRow,
    SessionRow,
    SessionStatus,
    list_schedulable_agents_by_sgroup,
    recalc_agent_resource_occupancy,
    recalc_concurrency_used,
)
from ..models.utils import ExtendedAsyncSAEngine as SAEngine
from ..models.utils import execute_with_retry, sql_json_increment, sql_json_merge
from .predicates import (
    check_concurrency,
    check_dependencies,
    check_domain_resource_limit,
    check_group_resource_limit,
    check_keypair_resource_limit,
    check_pending_session_count_limit,
    check_pending_session_resource_limit,
    check_reserved_batch_session,
    check_user_resource_limit,
)
from .types import (
    AbstractScheduler,
    AgentAllocationContext,
    KernelAgentBinding,
    PendingSession,
    PredicateResult,
    SchedulingContext,
)

if TYPE_CHECKING:
    from ..config import LocalConfig, SharedConfig
    from ..registry import AgentRegistry

__all__ = (
    "load_scheduler",
    "SchedulerDispatcher",
)

log = BraceStyleAdapter(logging.getLogger("ai.backend.manager.scheduler"))

_log_fmt: ContextVar[str] = ContextVar("_log_fmt")
_log_args: ContextVar[Tuple[Any, ...]] = ContextVar("_log_args")

_key_schedule_prep_tasks: Final = "scheduler.preptasks"


def get_schedulable_group_id(agents: list[AgentRow]) -> str:
    return hashlib.md5("#".join(list(map(lambda agent: agent.id, agents))).encode()).hexdigest()


def load_scheduler(
    name: str,
    sgroup_opts: ScalingGroupOpts,
    scheduler_config: dict[str, Any],
) -> AbstractScheduler:
    entry_prefix = "backendai_scheduler_v10"
    for entrypoint in scan_entrypoints(entry_prefix):
        if entrypoint.name == name:
            log.debug('loading scheduler plugin "{}" from {}', name, entrypoint.module)
            scheduler_cls = entrypoint.load()
            return scheduler_cls(sgroup_opts, scheduler_config)
    raise ImportError("Cannot load the scheduler plugin", name)


StartTaskArgs = Tuple[
    Tuple[Any, ...],
    SchedulingContext,
    Tuple[PendingSession, List[KernelAgentBinding]],
    List[Tuple[str, Union[Exception, PredicateResult]]],
]


class SchedulerDispatcher(aobject):
    config: LocalConfig
    shared_config: SharedConfig
    registry: AgentRegistry
    db: SAEngine

    event_dispatcher: EventDispatcher
    event_producer: EventProducer
    schedule_timer: GlobalTimer
    prepare_timer: GlobalTimer
    scale_timer: GlobalTimer

    redis_live: RedisConnectionInfo

    def __init__(
        self,
        local_config: LocalConfig,
        shared_config: SharedConfig,
        event_dispatcher: EventDispatcher,
        event_producer: EventProducer,
        lock_factory: DistributedLockFactory,
        registry: AgentRegistry,
    ) -> None:
        self.local_config = local_config
        self.shared_config = shared_config
        self.event_dispatcher = event_dispatcher
        self.event_producer = event_producer
        self.registry = registry
        self.lock_factory = lock_factory
        self.db = registry.db
        self.redis_live = redis_helper.get_redis_object(
            self.shared_config.data["redis"],
            name="scheduler.live",
            db=REDIS_LIVE_DB,
        )

    async def __ainit__(self) -> None:
        coalescing_opts: CoalescingOptions = {
            "max_wait": 0.5,
            "max_batch_size": 32,
        }
        # coalescing_opts = None
        evd = self.registry.event_dispatcher
        evd.consume(
            SessionEnqueuedEvent, None, self.schedule, coalescing_opts, name="dispatcher.enq"
        )
        evd.consume(
            SessionTerminatedEvent, None, self.schedule, coalescing_opts, name="dispatcher.term"
        )
        evd.consume(AgentStartedEvent, None, self.schedule)
        evd.consume(DoScheduleEvent, None, self.schedule, coalescing_opts)
        evd.consume(DoPrepareEvent, None, self.prepare)
        evd.consume(DoScaleEvent, None, self.scale_services)
        self.schedule_timer = GlobalTimer(
            self.lock_factory(LockID.LOCKID_SCHEDULE_TIMER, 10.0),
            self.event_producer,
            lambda: DoScheduleEvent(),
            interval=10.0,
            task_name="schedule_timer",
        )
        self.prepare_timer = GlobalTimer(
            self.lock_factory(LockID.LOCKID_PREPARE_TIMER, 10.0),
            self.event_producer,
            lambda: DoPrepareEvent(),
            interval=10.0,
            initial_delay=5.0,
            task_name="prepare_timer",
        )
        self.scale_timer = GlobalTimer(
            self.lock_factory(LockID.LOCKID_SCALE_TIMER, 10.0),
            self.event_producer,
            lambda: DoScaleEvent(),
            interval=10.0,
            initial_delay=7.0,
            task_name="scale_timer",
        )
        await self.schedule_timer.join()
        await self.prepare_timer.join()
        await self.scale_timer.join()
        log.info("Session scheduler started")

    async def close(self) -> None:
        async with aiotools.TaskGroup() as tg:
            tg.create_task(self.scale_timer.leave())
            tg.create_task(self.prepare_timer.leave())
            tg.create_task(self.schedule_timer.leave())
        await self.redis_live.close()
        log.info("Session scheduler stopped")

    async def schedule(
        self,
        context: None,
        source: AgentId,
        event: SessionEnqueuedEvent | SessionTerminatedEvent | AgentStartedEvent | DoScheduleEvent,
    ) -> None:
        """
        Trigger the scheduler to scan pending sessions and mark them scheduled if they fulfill
        the scheduling requirements.

        HoL blocking issue due to indefinitely preparing sessions will be mitigated because
        they will be treated as already "scheduled" sessions and the scheduler will continue to
        work on other pending sessions.

        Session status transition: PENDING -> SCHEDULED
        """
        log.debug("schedule(): triggered")
        manager_id = self.local_config["manager"]["id"]
        redis_key = f"manager.{manager_id}.schedule"

        def _pipeline(r: Redis) -> RedisPipeline:
            pipe = r.pipeline()
            pipe.delete(redis_key)
            pipe.hset(
                redis_key,
                mapping={
                    "trigger_event": event.__class__.name,
                    "execution_time": datetime.now(tzutc()).isoformat(),
                },
            )
            return pipe

        await redis_helper.execute(
            self.redis_live,
            _pipeline,
        )
        known_slot_types = await self.shared_config.get_resource_slots()
        sched_ctx = SchedulingContext(
            registry=self.registry,
            known_slot_types=known_slot_types,
        )

        try:
            # The schedule() method should be executed with a global lock
            # as its individual steps are composed of many short-lived transactions.
            async with self.lock_factory(LockID.LOCKID_SCHEDULE, 60):
                async with self.db.begin_readonly_session() as db_sess:
                    # query = (
                    #     sa.select(ScalingGroupRow)
                    #     .join(ScalingGroupRow.agents.and_(AgentRow.status == AgentStatus.ALIVE))
                    # )
                    query = (
                        sa.select(AgentRow.scaling_group)
                        .where(AgentRow.status == AgentStatus.ALIVE)
                        .group_by(AgentRow.scaling_group)
                    )
                    result = await db_sess.execute(query)
                    schedulable_scaling_groups = [row.scaling_group for row in result.fetchall()]
                for sgroup_name in schedulable_scaling_groups:
                    try:
                        await self._schedule_in_sgroup(
                            sched_ctx,
                            sgroup_name,
                        )
                        await redis_helper.execute(
                            self.redis_live,
                            lambda r: r.hset(
                                redis_key,
                                "resource_group",
                                sgroup_name,
                            ),
                        )
                    except InstanceNotAvailable as e:
                        # Proceed to the next scaling group and come back later.
                        log.debug(
                            "schedule({}): instance not available ({})",
                            sgroup_name,
                            e.extra_msg,
                        )
                    except Exception as e:
                        log.exception("schedule({}): scheduling error!\n{}", sgroup_name, repr(e))
                await redis_helper.execute(
                    self.redis_live,
                    lambda r: r.hset(
                        redis_key,
                        "finish_time",
                        datetime.now(tzutc()).isoformat(),
                    ),
                )
        except DBAPIError as e:
            if getattr(e.orig, "pgcode", None) == "55P03":
                log.info(
                    "schedule(): cancelled due to advisory lock timeout; "
                    "maybe another schedule() call is still running"
                )
                raise asyncio.CancelledError()
            raise

    async def _load_scheduler(
        self,
        db_sess: SASession,
        sgroup_name: str,
    ) -> AbstractScheduler:
        query = sa.select(ScalingGroupRow.scheduler, ScalingGroupRow.scheduler_opts).where(
            ScalingGroupRow.name == sgroup_name
        )
        result = await db_sess.execute(query)
        row = result.first()
        scheduler_name = row.scheduler
        sgroup_opts: ScalingGroupOpts = row.scheduler_opts
        global_scheduler_opts = {}
        if self.shared_config["plugins"]["scheduler"]:
            global_scheduler_opts = self.shared_config["plugins"]["scheduler"].get(
                scheduler_name, {}
            )
        scheduler_specific_config = {**global_scheduler_opts, **sgroup_opts.config}
        return load_scheduler(scheduler_name, sgroup_opts, scheduler_specific_config)

    async def _schedule_in_sgroup(
        self,
        sched_ctx: SchedulingContext,
        sgroup_name: str,
    ) -> None:
        async def _apply_cancellation(
            db_sess: SASession, session_ids: list[SessionId], reason="pending-timeout"
        ):
            now = datetime.now(tzutc())
            kernel_query = (
                sa.update(KernelRow)
                .values(
                    status=KernelStatus.CANCELLED,
                    status_info=reason,
                    terminated_at=now,
                    status_history=sql_json_merge(
                        KernelRow.status_history,
                        (),
                        {
                            KernelStatus.CANCELLED.name: now.isoformat(),
                        },
                    ),
                )
                .where(KernelRow.session_id.in_(session_ids))
            )
            await db_sess.execute(kernel_query)
            query = (
                sa.update(SessionRow)
                .values(
                    status=SessionStatus.CANCELLED,
                    status_info=reason,
                    terminated_at=now,
                    status_history=sql_json_merge(
                        SessionRow.status_history,
                        (),
                        {
                            SessionStatus.CANCELLED.name: now.isoformat(),
                        },
                    ),
                )
                .where(SessionRow.id.in_(session_ids))
            )
            await db_sess.execute(query)

        async with self.db.begin_readonly_session() as db_sess:
            scheduler = await self._load_scheduler(db_sess, sgroup_name)
            existing_sessions, pending_sessions, cancelled_sessions = await _list_managed_sessions(
                db_sess, sgroup_name, scheduler.sgroup_opts.pending_timeout
            )

        if cancelled_sessions:
            session_ids = [item.id for item in cancelled_sessions]

            async def _update():
                async with self.db.begin_session() as db_sess:
                    await _apply_cancellation(db_sess, session_ids)

            await execute_with_retry(_update)
            for item in cancelled_sessions:
                await self.event_producer.produce_event(
                    SessionCancelledEvent(
                        item.id,
                        item.creation_id,
                        reason=KernelLifecycleEventReason.PENDING_TIMEOUT,
                    ),
                )

        log.debug(
            "running scheduler (sgroup:{}, pending:{}, existing:{}, cancelled:{})",
            sgroup_name,
            len(pending_sessions),
            len(existing_sessions),
            len(cancelled_sessions),
        )
        zero = ResourceSlot()
        num_scheduled = 0
        while len(pending_sessions) > 0:
            async with self.db.begin_readonly_session() as db_sess:
                candidate_agents = await list_schedulable_agents_by_sgroup(db_sess, sgroup_name)
            total_capacity = sum((ag.available_slots for ag in candidate_agents), zero)

            picked_session_id = scheduler.pick_session(
                total_capacity,
                pending_sessions,
                existing_sessions,
            )
            if picked_session_id is None:
                # no session is picked.
                # continue to next sgroup.
                return
            for picked_idx, sess_ctx in enumerate(pending_sessions):
                if sess_ctx.id == picked_session_id:
                    break
            else:
                # no matching entry for picked session?
                raise RuntimeError("should not reach here")
            sess_ctx = pending_sessions.pop(picked_idx)
            log_fmt = "schedule(s:{}, type:{}, name:{}, ak:{}, cluster_mode:{}): "
            log_args = (
                sess_ctx.id,
                sess_ctx.session_type,
                sess_ctx.name,
                sess_ctx.access_key,
                sess_ctx.cluster_mode,
            )
            _log_fmt.set(log_fmt)
            _log_args.set(log_args)
            log.debug(log_fmt + "try-scheduling", *log_args)

            async def _check_predicates() -> List[Tuple[str, Union[Exception, PredicateResult]]]:
                check_results: List[Tuple[str, Union[Exception, PredicateResult]]] = []
                async with self.db.begin_session() as db_sess:
                    predicates: list[Tuple[str, Awaitable[PredicateResult]]] = [
                        (
                            "reserved_time",
                            check_reserved_batch_session(db_sess, sched_ctx, sess_ctx),
                        ),
                        ("dependencies", check_dependencies(db_sess, sched_ctx, sess_ctx)),
                        ("concurrency", check_concurrency(db_sess, sched_ctx, sess_ctx)),
                    ]
                    if not sess_ctx.is_private:
                        predicates += [
                            (
                                "pending_session_resource_limit",
                                check_pending_session_resource_limit(db_sess, sched_ctx, sess_ctx),
                            ),
                            (
                                "pending_session_count_limit",
                                check_pending_session_count_limit(db_sess, sched_ctx, sess_ctx),
                            ),
                            (
                                "keypair_resource_limit",
                                check_keypair_resource_limit(db_sess, sched_ctx, sess_ctx),
                            ),
                            (
                                "user_resource_limit",
                                check_user_resource_limit(db_sess, sched_ctx, sess_ctx),
                            ),
                            (
                                "user_group_resource_limit",
                                check_group_resource_limit(db_sess, sched_ctx, sess_ctx),
                            ),
                            (
                                "domain_resource_limit",
                                check_domain_resource_limit(db_sess, sched_ctx, sess_ctx),
                            ),
                        ]
                    for predicate_name, check_coro in predicates:
                        try:
                            check_results.append((predicate_name, await check_coro))
                        except DBAPIError:
                            raise
                        except Exception as e:
                            log.exception(log_fmt + "predicate-error", *log_args)
                            check_results.append((predicate_name, e))
                return check_results

            check_results = await execute_with_retry(_check_predicates)
            failed_predicates = []
            passed_predicates = []
            for predicate_name, result in check_results:
                if isinstance(result, Exception):
                    failed_predicates.append({
                        "name": predicate_name,
                        "msg": repr(result),
                    })
                    continue
                if result.passed:
                    passed_predicates.append({
                        "name": predicate_name,
                    })
                else:
                    failed_predicates.append({
                        "name": predicate_name,
                        "msg": result.message or "",
                    })

            async def _check_predicates_hook() -> HookResult:
                async with self.db.begin_readonly_session() as db_sess:
                    return await self.registry.hook_plugin_ctx.dispatch(
                        "PREDICATE",
                        (
                            db_sess,
                            sched_ctx,
                            sess_ctx,
                        ),
                    )

            hook_result = await execute_with_retry(_check_predicates_hook)
            match hook_result.src_plugin:
                case str():
                    hook_name = hook_result.src_plugin
                case list():
                    hook_name = f"({', '.join(hook_result.src_plugin)})"
                case _:
                    hook_name = ""

            if hook_result.status == PASSED:
                if hook_result.src_plugin:
                    # Append result only when plugin exists.
                    passed_predicates.append({"name": hook_name})
            else:
                failed_predicates.append({
                    "name": hook_name,
                    "msg": hook_result.reason or "",
                })

            status_update_data = {
                "last_try": datetime.now(tzutc()).isoformat(),
                "failed_predicates": failed_predicates,
                "passed_predicates": passed_predicates,
            }
            if failed_predicates:
                log.debug(log_fmt + "predicate-checks-failed (temporary)", *log_args)

                async def _cancel_failed_system_session() -> None:
                    async with self.db.begin_session() as db_sess:
                        await _rollback_predicate_mutations(
                            db_sess,
                            sched_ctx,
                            sess_ctx,
                        )
                        query = (
                            sa.update(SessionRow)
                            .values(
                                status_info="predicate-checks-failed",
                                status_data=sql_json_increment(
                                    SessionRow.status_data,
                                    ("scheduler", "retries"),
                                    parent_updates=status_update_data,
                                ),
                            )
                            .where(SessionRow.id == sess_ctx.id)
                        )
                        await db_sess.execute(query)
                        if sess_ctx.is_private:
                            await _apply_cancellation(db_sess, [sess_ctx.id])
                            await self.event_producer.produce_event(
                                SessionCancelledEvent(
                                    sess_ctx.id,
                                    sess_ctx.creation_id,
                                    reason=KernelLifecycleEventReason.PENDING_TIMEOUT,
                                )
                            )

                await execute_with_retry(_cancel_failed_system_session)
                # Predicate failures are *NOT* permanent errors.
                # We need to retry the scheduling afterwards.
                continue
            else:

                async def _update_session_status_data() -> None:
                    async with self.db.begin_session() as db_sess:
                        kernel_query = (
                            sa.update(KernelRow)
                            .where(KernelRow.session_id == sess_ctx.id)
                            .values(
                                status_data=sql_json_merge(
                                    KernelRow.status_data,
                                    ("scheduler",),
                                    obj=status_update_data,
                                ),
                            )
                        )
                        await db_sess.execute(kernel_query)
                        session_query = (
                            sa.update(SessionRow)
                            .where(SessionRow.id == sess_ctx.id)
                            .values(
                                status_data=sql_json_merge(
                                    SessionRow.status_data,
                                    ("scheduler",),
                                    obj=status_update_data,
                                ),
                            )
                        )
                        await db_sess.execute(session_query)

                await execute_with_retry(_update_session_status_data)

            async with self.db.begin_readonly_session() as db_sess:
                schedulable_sess = await SessionRow.get_session_by_id(
                    db_sess,
                    sess_ctx.id,
                    eager_loading_op=(
                        noload("*"),
                        selectinload(SessionRow.kernels).options(
                            noload("*"),
                            selectinload(KernelRow.agent_row).noload("*"),
                        ),
                    ),
                )

            agent_selection_resource_priority = self.local_config["manager"][
                "agent-selection-resource-priority"
            ]

            if schedulable_sess.cluster_mode == ClusterMode.SINGLE_NODE:
                await self._schedule_single_node_session(
                    sched_ctx,
                    scheduler,
                    sgroup_name,
                    candidate_agents,
                    schedulable_sess,
                    agent_selection_resource_priority,
                    check_results,
                )
            elif schedulable_sess.cluster_mode == ClusterMode.MULTI_NODE:
                await self._schedule_multi_node_session(
                    sched_ctx,
                    scheduler,
                    sgroup_name,
                    candidate_agents,
                    schedulable_sess,
                    agent_selection_resource_priority,
                    check_results,
                )
            else:
                raise RuntimeError(
                    f"should not reach here; unknown cluster_mode: {schedulable_sess.cluster_mode}",
                )
            num_scheduled += 1
        if num_scheduled > 0:
            await self.event_producer.produce_event(DoPrepareEvent())

    async def _filter_agent_by_container_limit(
        self, candidate_agents: list[AgentRow]
    ) -> list[AgentRow]:
        raw_value = await self.shared_config.etcd.get("config/agent/max-container-count")
        if raw_value is None:
            return candidate_agents
        max_container_count = int(raw_value)

        async def _pipe_builder(r: Redis) -> RedisPipeline:
            pipe = r.pipeline()
            for ag in candidate_agents:
                await pipe.get(f"container_count.{ag.id}")
            return pipe

        raw_counts = await redis_helper.execute(self.registry.redis_stat, _pipe_builder)

        def _check(cnt: str | None) -> bool:
            _cnt = int(cnt) if cnt is not None else 0
            return max_container_count > _cnt

        return [ag for ag, count in zip(candidate_agents, raw_counts) if _check(count)]

    async def _schedule_single_node_session(
        self,
        sched_ctx: SchedulingContext,
        scheduler: AbstractScheduler,
        sgroup_name: str,
        candidate_agents: Sequence[AgentRow],
        sess_ctx: SessionRow,
        agent_selection_resource_priority: list[str],
        check_results: List[Tuple[str, Union[Exception, PredicateResult]]],
    ) -> None:
        """
        Finds and assigns an agent having resources enough to host the entire session.
        """
        log_fmt = _log_fmt.get("")
        log_args = _log_args.get(tuple())
        requested_architectures = set(k.architecture for k in sess_ctx.kernels)
        if len(requested_architectures) > 1:
            raise GenericBadRequest(
                "Cannot assign multiple kernels with different architectures' single node session",
            )
        if not sess_ctx.kernels:
            raise GenericBadRequest(f"The session {sess_ctx.id!r} does not have any child kernel.")
        requested_architecture = requested_architectures.pop()
        compatible_candidate_agents = [
            ag for ag in candidate_agents if ag.architecture == requested_architecture
        ]

        try:
            if not candidate_agents:
                raise InstanceNotAvailable(extra_msg="No agents are available for scheduling")
            if not compatible_candidate_agents:
                raise InstanceNotAvailable(
                    extra_msg=(
                        "No agents found to be compatible with the image architecture "
                        f"(image[0]: {sess_ctx.main_kernel.image_ref}, "
                        f"arch: {requested_architecture})"
                    ),
                )
            available_candidate_agents = await self._filter_agent_by_container_limit(
                compatible_candidate_agents
            )
            if not available_candidate_agents:
                raise InstanceNotAvailable(
                    extra_msg=(
                        "No agents found to be available because all agents have reached the hard"
                        " limit of the number of containers."
                    ),
                )

            # If sess_ctx.agent_id is already set for manual assignment by superadmin,
            # skip assign_agent_for_session().
            agent = sess_ctx.main_kernel.agent_row
            agent_id: AgentId | None = None
            if agent is not None:
                agent_id = agent.id

                async with self.db.begin_session() as db_sess:
                    result = (
                        await db_sess.execute(
                            sa.select([AgentRow.available_slots, AgentRow.occupied_slots]).where(
                                AgentRow.id == agent_id
                            )
                        )
                    ).fetchall()[0]

                if result is None:
                    raise GenericBadRequest(f"No such agent exist in DB: {agent_id}")

                available_slots, occupied_slots = result

                for key in available_slots.keys():
                    if (
                        available_slots[key] - occupied_slots.get(key, Decimal("0"))
                        >= sess_ctx.requested_slots[key]
                    ):
                        continue
                    else:
                        raise InstanceNotAvailable(
                            extra_msg=(
                                f"The designated agent ({agent_id}) does not have "
                                f"the enough remaining capacity ({key}, "
                                f"requested: {sess_ctx.requested_slots[key]}, "
                                f"remaining: {available_slots[key] - occupied_slots[key]})."
                            ),
                        )

            else:
                sorted_agents = sorted(compatible_candidate_agents, key=lambda agent: agent.id)

                if scheduler.sgroup_opts.roundrobin:
                    rr_state: (
                        RoundRobinState | None
                    ) = await sched_ctx.registry.shared_config.get_roundrobin_state(
                        sgroup_name, requested_architecture
                    )

                    if rr_state is not None:
                        schedulable_group_id = get_schedulable_group_id(sorted_agents)

                        if schedulable_group_id == rr_state.schedulable_group_id:
                            for i in range(len(sorted_agents)):
                                idx = (rr_state.next_index + i) % len(sorted_agents)
                                agent = sorted_agents[idx]

                                if (
                                    agent.available_slots - agent.occupied_slots
                                    > sess_ctx.requested_slots
                                ):
                                    agent_id = agent.id
                                    rr_state.next_index = (rr_state.next_index + i + 1) % len(
                                        sorted_agents
                                    )

                                    await sched_ctx.registry.shared_config.put_roundrobin_state(
                                        sgroup_name, requested_architecture, rr_state
                                    )
                                    break
                            else:
                                # fallback to the default behavior instead of raising an error for reducing code complexity
                                pass

                if agent_id is None:
                    # Let the scheduler check the resource availability and decide the target agent
                    cand_agent_id = scheduler.assign_agent_for_session(
                        compatible_candidate_agents,
                        sess_ctx,
                        scheduler.sgroup_opts.agent_selection_strategy,
                        agent_selection_resource_priority,
                    )
                    if cand_agent_id is None:
                        raise InstanceNotAvailable(
                            extra_msg=(
                                "Could not find a contiguous resource region in any agent big"
                                f" enough to host the session (id: {sess_ctx.id}, resource group:"
                                f" {sess_ctx.scaling_group_name})"
                            ),
                        )
                    agent_id = cand_agent_id

                    if scheduler.sgroup_opts.roundrobin:
                        await sched_ctx.registry.shared_config.put_roundrobin_state(
                            sgroup_name,
                            requested_architecture,
                            RoundRobinState(
                                schedulable_group_id=get_schedulable_group_id(
                                    sorted_agents,
                                ),
                                next_index=[
                                    (idx + 1) % len(sorted_agents)
                                    for idx, agent in enumerate(sorted_agents)
                                    if agent.id == agent_id
                                ][0],
                            ),
                        )

            async with self.db.begin_session() as agent_db_sess:
                agent_alloc_ctx = await _reserve_agent(
                    sched_ctx,
                    agent_db_sess,
                    sgroup_name,
                    agent_id,
                    sess_ctx.requested_slots,
                )
        except InstanceNotAvailable as sched_failure:
            log.debug(log_fmt + "no-available-instances", *log_args)

            async def _update_sched_failure(exc: InstanceNotAvailable) -> None:
                async with self.db.begin_session() as kernel_db_sess:
                    await _rollback_predicate_mutations(
                        kernel_db_sess,
                        sched_ctx,
                        sess_ctx,
                    )
                    query = (
                        sa.update(SessionRow)
                        .values(
                            status_info="no-available-instances",
                            status_data=sql_json_increment(
                                SessionRow.status_data,
                                ("scheduler", "retries"),
                                parent_updates={
                                    "last_try": datetime.now(tzutc()).isoformat(),
                                    "msg": exc.extra_msg,
                                },
                            ),
                        )
                        .where(SessionRow.id == sess_ctx.id)
                    )
                    await kernel_db_sess.execute(query)

            await execute_with_retry(partial(_update_sched_failure, sched_failure))
            raise
        except Exception as e:
            log.exception(
                log_fmt + "unexpected-error, during agent allocation",
                *log_args,
            )
            exc_data = convert_to_status_data(e, self.local_config["debug"]["enabled"])

            async def _update_generic_failure() -> None:
                async with self.db.begin_session() as kernel_db_sess:
                    await _rollback_predicate_mutations(
                        kernel_db_sess,
                        sched_ctx,
                        sess_ctx,
                    )
                    query = (
                        sa.update(SessionRow)
                        .values(
                            status_info="scheduler-error",
                            status_data=exc_data,
                        )
                        .where(SessionRow.id == sess_ctx.id)
                    )
                    await kernel_db_sess.execute(query)

            await execute_with_retry(_update_generic_failure)
            raise

        async def _finalize_scheduled() -> None:
            agent_ids: list[AgentId] = []
            async with self.db.begin_session() as db_sess:
                now = datetime.now(tzutc())
                for kernel in sess_ctx.kernels:
                    kernel_query = (
                        sa.update(KernelRow)
                        .values(
                            agent=agent_alloc_ctx.agent_id,
                            agent_addr=agent_alloc_ctx.agent_addr,
                            scaling_group=sgroup_name,
                            status=KernelStatus.SCHEDULED,
                            status_info="scheduled",
                            status_data={},
                            status_changed=now,
                            status_history=sql_json_merge(
                                KernelRow.status_history,
                                (),
                                {
                                    KernelStatus.SCHEDULED.name: now.isoformat(),
                                },
                            ),
                        )
                        .where(KernelRow.id == kernel.id)
                    )
                    await db_sess.execute(kernel_query)
                if agent_alloc_ctx.agent_id is not None:
                    agent_ids.append(agent_alloc_ctx.agent_id)

                session_query = (
                    sa.update(SessionRow)
                    .values(
                        scaling_group_name=sgroup_name,
                        agent_ids=agent_ids,
                        status=SessionStatus.SCHEDULED,
                        status_info="scheduled",
                        status_data={},
                        status_history=sql_json_merge(
                            SessionRow.status_history,
                            (),
                            {
                                SessionStatus.SCHEDULED.name: now.isoformat(),
                            },
                        ),
                    )
                    .where(SessionRow.id == sess_ctx.id)
                )
                await db_sess.execute(session_query)

        await execute_with_retry(_finalize_scheduled)
        await self.registry.event_producer.produce_event(
            SessionScheduledEvent(sess_ctx.id, sess_ctx.creation_id),
        )

    async def _schedule_multi_node_session(
        self,
        sched_ctx: SchedulingContext,
        scheduler: AbstractScheduler,
        sgroup_name: str,
        candidate_agents: Sequence[AgentRow],
        sess_ctx: SessionRow,
        agent_selection_resource_priority: list[str],
        check_results: List[Tuple[str, Union[Exception, PredicateResult]]],
    ) -> None:
        """
        Finds and assigns agents having resources enough to host each kernel in the session.
        """
        log_fmt = _log_fmt.get()
        log_args = _log_args.get()
        agent_query_extra_conds = None
        kernel_agent_bindings: List[KernelAgentBinding] = []
        async with self.db.begin_session() as agent_db_sess:
            # This outer transaction is rolled back when any exception occurs inside,
            # including scheduling failures of a kernel.
            # It ensures that occupied_slots are recovered when there are partial
            # scheduling failures.
            kernel: KernelRow
            for kernel in sess_ctx.kernels:
                agent_alloc_ctx: AgentAllocationContext | None = None
                try:
                    agent_id: Optional[AgentId] = None
                    agent: Optional[AgentRow] = kernel.agent_row
                    if agent is not None:
                        # Check the resource availability of the manually designated agent
                        result = (
                            await agent_db_sess.execute(
                                sa.select([
                                    AgentRow.available_slots,
                                    AgentRow.occupied_slots,
                                ]).where(AgentRow.id == agent.id)
                            )
                        ).fetchall()[0]

                        if result is None:
                            raise GenericBadRequest(f"No such agent exist in DB: {agent_id}")
                        available_slots, occupied_slots = result

                        for key in available_slots.keys():
                            if (
                                available_slots[key] - occupied_slots[key]
                                >= kernel.requested_slots[key]
                            ):
                                continue
                            else:
                                raise InstanceNotAvailable(
                                    extra_msg=(
                                        f"The designated agent ({agent.id}) does not have "
                                        f"the enough remaining capacity ({key}, "
                                        f"requested: {sess_ctx.requested_slots[key]}, "
                                        f"remaining: {available_slots[key] - occupied_slots[key]})."
                                    ),
                                )
                        agent_id = agent.id
                    else:
                        # Each kernel may have different images and different architectures
                        compatible_candidate_agents = [
                            ag for ag in candidate_agents if ag.architecture == kernel.architecture
                        ]
                        if not candidate_agents:
                            raise InstanceNotAvailable(
                                extra_msg="No agents are available for scheduling"
                            )
                        if not compatible_candidate_agents:
                            raise InstanceNotAvailable(
                                extra_msg=(
                                    "No agents found to be compatible with the image architecture "
                                    f"(image: {kernel.image_ref}, "
                                    f"arch: {kernel.architecture})"
                                ),
                            )
                        available_candidate_agents = await self._filter_agent_by_container_limit(
                            compatible_candidate_agents
                        )
                        if not available_candidate_agents:
                            raise InstanceNotAvailable(
                                extra_msg=(
                                    "No agents found to be available because all agents have"
                                    " reached the hard limit of the number of containers."
                                ),
                            )
                        # Let the scheduler check the resource availability and decide the target agent
                        agent_id = scheduler.assign_agent_for_kernel(
                            available_candidate_agents,
                            kernel,
                            scheduler.sgroup_opts.agent_selection_strategy,
                            agent_selection_resource_priority,
                        )
                        if agent_id is None:
                            raise InstanceNotAvailable(
                                extra_msg=(
                                    "Could not find a contiguous resource region in any agent big"
                                    f" enough to host a kernel in the session (id: {sess_ctx.id},"
                                    f" resource group: {sess_ctx.scaling_group_name})"
                                ),
                            )
                    assert agent_id is not None

                    async def _reserve() -> None:
                        nonlocal agent_alloc_ctx, candidate_agents
                        async with agent_db_sess.begin_nested():
                            agent_alloc_ctx = await _reserve_agent(
                                sched_ctx,
                                agent_db_sess,
                                sgroup_name,
                                agent_id,
                                kernel.requested_slots,
                                extra_conds=agent_query_extra_conds,
                            )
                            # Update the agent data to schedule the next kernel in the session
                            candidate_agents = await list_schedulable_agents_by_sgroup(
                                agent_db_sess,
                                sgroup_name,
                            )

                    await execute_with_retry(_reserve)
                except InstanceNotAvailable as sched_failure:
                    log.debug(log_fmt + "no-available-instances", *log_args)

                    async def _update_sched_failure(exc: InstanceNotAvailable) -> None:
                        async with self.db.begin_session() as agent_db_sess:
                            await _rollback_predicate_mutations(
                                agent_db_sess,
                                sched_ctx,
                                sess_ctx,
                            )
                            query = (
                                sa.update(KernelRow)
                                .values(
                                    status_info="no-available-instances",
                                    status_data=sql_json_increment(
                                        KernelRow.status_data,
                                        ("scheduler", "retries"),
                                        parent_updates={
                                            "last_try": datetime.now(tzutc()).isoformat(),
                                            "msg": exc.extra_msg,
                                        },
                                    ),
                                )
                                .where(KernelRow.id == kernel.id)
                            )
                            await agent_db_sess.execute(query)

                    await execute_with_retry(partial(_update_sched_failure, sched_failure))
                    raise
                except Exception as e:
                    log.exception(
                        log_fmt + "unexpected-error, during agent allocation",
                        *log_args,
                    )
                    exc_data = convert_to_status_data(e, self.local_config["debug"]["enabled"])

                    async def _update_generic_failure() -> None:
                        async with self.db.begin_session() as kernel_db_sess:
                            await _rollback_predicate_mutations(
                                kernel_db_sess,
                                sched_ctx,
                                sess_ctx,
                            )
                            query = (
                                sa.update(KernelRow)
                                .values(
                                    status_info="scheduler-error",
                                    status_data=exc_data,
                                )
                                .where(KernelRow.id == kernel.id)
                            )
                            await kernel_db_sess.execute(query)

                    await execute_with_retry(_update_generic_failure)
                    raise
                else:
                    assert agent_alloc_ctx is not None
                    kernel_agent_bindings.append(KernelAgentBinding(kernel, agent_alloc_ctx, set()))

        assert len(kernel_agent_bindings) == len(sess_ctx.kernels)
        # Proceed to PREPARING only when all kernels are successfully scheduled.

        async def _finalize_scheduled() -> None:
            agent_ids: list[AgentId] = []
            async with self.db.begin_session() as db_sess:
                for binding in kernel_agent_bindings:
                    now = datetime.now(tzutc())
                    kernel_query = (
                        sa.update(KernelRow)
                        .values(
                            agent=binding.agent_alloc_ctx.agent_id,
                            agent_addr=binding.agent_alloc_ctx.agent_addr,
                            scaling_group=sgroup_name,
                            status=KernelStatus.SCHEDULED,
                            status_info="scheduled",
                            status_data={},
                            status_changed=now,
                            status_history=sql_json_merge(
                                KernelRow.status_history,
                                (),
                                {
                                    KernelStatus.SCHEDULED.name: now.isoformat(),
                                },
                            ),
                        )
                        .where(KernelRow.id == binding.kernel.id)
                    )
                    await db_sess.execute(kernel_query)
                    if binding.agent_alloc_ctx.agent_id is not None:
                        agent_ids.append(binding.agent_alloc_ctx.agent_id)

                session_query = (
                    sa.update(SessionRow)
                    .values(
                        scaling_group_name=sgroup_name,
                        agent_ids=agent_ids,
                        status=SessionStatus.SCHEDULED,
                        status_info="scheduled",
                        status_data={},
                        # status_changed=now,
                        status_history=sql_json_merge(
                            SessionRow.status_history,
                            (),
                            {
                                SessionStatus.SCHEDULED.name: now.isoformat(),
                            },
                        ),
                    )
                    .where(SessionRow.id == sess_ctx.id)
                )
                await db_sess.execute(session_query)

        await execute_with_retry(_finalize_scheduled)
        await self.registry.event_producer.produce_event(
            SessionScheduledEvent(sess_ctx.id, sess_ctx.creation_id),
        )

    async def prepare(
        self,
        context: None,
        source: AgentId,
        event: DoPrepareEvent,
    ) -> None:
        """
        Scan the scheduled sessions and perform the agent RPC calls to begin preparation of them.
        Each RPC calls are done in separate asyncio tasks.

        Session status transition: SCHEDULED -> PREPARING
        """
        manager_id = self.local_config["manager"]["id"]
        redis_key = f"manager.{manager_id}.prepare"

        def _pipeline(r: Redis) -> RedisPipeline:
            pipe = r.pipeline()
            pipe.delete(redis_key)
            pipe.hset(
                redis_key,
                mapping={
                    "trigger_event": event.__class__.name,
                    "execution_time": datetime.now(tzutc()).isoformat(),
                },
            )
            return pipe

        await redis_helper.execute(
            self.redis_live,
            _pipeline,
        )
        known_slot_types = await self.shared_config.get_resource_slots()
        sched_ctx = SchedulingContext(
            self.registry,
            known_slot_types,
        )
        try:
            async with self.lock_factory(LockID.LOCKID_PREPARE, 600):
                now = datetime.now(tzutc())

                async def _mark_session_preparing() -> Sequence[SessionRow]:
                    async with self.db.begin_session() as db_sess:
                        update_query = (
                            sa.update(KernelRow)
                            .values(
                                status=KernelStatus.PREPARING,
                                status_changed=now,
                                status_info="",
                                status_data={},
                                status_history=sql_json_merge(
                                    KernelRow.status_history,
                                    (),
                                    {
                                        KernelStatus.PREPARING.name: now.isoformat(),
                                    },
                                ),
                            )
                            .where(
                                (KernelRow.status == KernelStatus.SCHEDULED),
                            )
                        )
                        await db_sess.execute(update_query)
                        update_sess_query = (
                            sa.update(SessionRow)
                            .values(
                                status=SessionStatus.PREPARING,
                                # status_changed=now,
                                status_info="",
                                status_data={},
                                status_history=sql_json_merge(
                                    SessionRow.status_history,
                                    (),
                                    {
                                        SessionStatus.PREPARING.name: now.isoformat(),
                                    },
                                ),
                            )
                            .where(SessionRow.status == SessionStatus.SCHEDULED)
                            .returning(SessionRow.id)
                        )
                        rows = (await db_sess.execute(update_sess_query)).fetchall()
                        if len(rows) == 0:
                            return []
                        target_session_ids = [r["id"] for r in rows]
                        select_query = (
                            sa.select(SessionRow)
                            .where(SessionRow.id.in_(target_session_ids))
                            .options(
                                noload("*"),
                                selectinload(SessionRow.kernels).noload("*"),
                            )
                        )
                        result = await db_sess.execute(select_query)
                        return result.scalars().all()

                scheduled_sessions: Sequence[SessionRow]
                scheduled_sessions = await execute_with_retry(_mark_session_preparing)
                log.debug("prepare(): preparing {} session(s)", len(scheduled_sessions))
                async with (
                    async_timeout.timeout(delay=50.0),
                    aiotools.PersistentTaskGroup() as tg,
                ):
                    for scheduled_session in scheduled_sessions:
                        await self.registry.event_producer.produce_event(
                            SessionPreparingEvent(
                                scheduled_session.id,
                                scheduled_session.creation_id,
                            ),
                        )
                        tg.create_task(
                            self.start_session(
                                sched_ctx,
                                scheduled_session,
                            )
                        )

                        await redis_helper.execute(
                            self.redis_live,
                            lambda r: r.hset(
                                redis_key, "resource_group", scheduled_session.scaling_group_name
                            ),
                        )
            await redis_helper.execute(
                self.redis_live,
                lambda r: r.hset(
                    redis_key,
                    "finish_time",
                    datetime.now(tzutc()).isoformat(),
                ),
            )
        except DBAPIError as e:
            if getattr(e.orig, "pgcode", None) == "55P03":
                log.info(
                    "prepare(): cancelled due to advisory lock timeout; "
                    "maybe another prepare() call is still running"
                )
                raise asyncio.CancelledError()
            raise
        except asyncio.TimeoutError:
            log.warning("prepare(): timeout while executing start_session()")

    async def scale_services(
        self,
        context: None,
        source: AgentId,
        event: DoScaleEvent,
    ) -> None:
        log.debug("scale_services(): triggered")
        # Altering inference sessions should only be done by invoking this method
        manager_id = self.local_config["manager"]["id"]
        redis_key = f"manager.{manager_id}.scale_services"

        def _pipeline(r: Redis) -> RedisPipeline:
            pipe = r.pipeline()
            pipe.delete(redis_key)
            pipe.hset(
                redis_key,
                mapping={
                    "trigger_event": event.__class__.name,
                    "execution_time": datetime.now(tzutc()).isoformat(),
                },
            )
            return pipe

        await redis_helper.execute(
            self.redis_live,
            _pipeline,
        )

        routes_to_destroy = []
        endpoints_to_expand: dict[EndpointRow, Any] = {}
        endpoints_to_mark_terminated: set[EndpointRow] = set()
        async with self.db.begin_readonly_session() as session:
            endpoints = await EndpointRow.list(
                session,
                load_image=True,
                load_routes=True,
                status_filter=[EndpointLifecycle.CREATED, EndpointLifecycle.DESTROYING],
            )
        for endpoint in endpoints:
            active_routings = [
                r for r in endpoint.routings if r.status != RouteStatus.FAILED_TO_START
            ]
            desired_session_count = endpoint.desired_session_count
            if (
                endpoint.lifecycle_stage == EndpointLifecycle.DESTROYING
                and len(active_routings) == 0
            ):
                endpoints_to_mark_terminated.add(endpoint)
                continue

            if len(active_routings) > desired_session_count:
                # We need to scale down!
                destroy_count = len(active_routings) - desired_session_count
                routes_to_destroy += list(
                    sorted(
                        [
                            route
                            for route in active_routings
                            if (
                                route.status != RouteStatus.PROVISIONING
                                and route.status != RouteStatus.TERMINATING
                            )
                        ],
                        key=lambda r: r.status == RouteStatus.UNHEALTHY,
                    )
                )[:destroy_count]
                log.debug(
                    "Shrinking {} from {} to {}",
                    endpoint.name,
                    len(active_routings),
                    endpoint.desired_session_count,
                )
            elif len(active_routings) < desired_session_count:
                if endpoint.retries > SERVICE_MAX_RETRIES:
                    continue
                # We need to scale up!
                create_count = desired_session_count - len(active_routings)
                endpoints_to_expand[endpoint] = create_count
                log.debug(
                    "Expanding {} from {} to {}",
                    endpoint.name,
                    len(active_routings),
                    endpoint.desired_session_count,
                )

        async with self.db.begin_readonly_session() as db_session:
            ids_of_session_to_destroy = [r.session for r in routes_to_destroy]
            kernel_loading_op = (
                noload("*"),
                selectinload(SessionRow.kernels).options(
                    noload("*"),
                    selectinload(KernelRow.agent_row).noload("*"),
                ),
            )
            query = _build_session_fetch_query(
                SessionRow.id.in_(ids_of_session_to_destroy), eager_loading_op=kernel_loading_op
            )
            result = await db_session.execute(query)
            target_sessions_to_destroy = result.scalars().all()

        already_destroyed_sessions: list[SessionId] = []
        # TODO: Update logic to not to wait for sessions to actually terminate
        for session in target_sessions_to_destroy:
            try:
                await self.registry.destroy_session(
                    session,
                    forced=True,
                    reason=KernelLifecycleEventReason.SERVICE_SCALED_DOWN,
                )
            except SessionNotFound:
                # Session already terminated while leaving routing alive
                already_destroyed_sessions.append(session.id)
        await redis_helper.execute(
            self.redis_live,
            lambda r: r.hset(
                redis_key,
                "down",
                json.dumps([str(s.id) for s in target_sessions_to_destroy]),
            ),
        )

        created_routes = []
        async with self.db.begin_session() as db_sess:
            for endpoint, expand_count in endpoints_to_expand.items():
                log.debug("Creating {} session(s) for {}", expand_count, endpoint.name)
                for _ in range(expand_count):
                    route_id = uuid.uuid4()
                    routing_row = RoutingRow(
                        route_id,
                        endpoint.id,
                        None,
                        endpoint.session_owner,
                        endpoint.domain,
                        endpoint.project,
                    )
                    db_sess.add(routing_row)
                    created_routes.append(route_id)
            await db_sess.commit()
        for route_id in created_routes:
            await self.event_producer.produce_event(RouteCreatedEvent(route_id))
        await redis_helper.execute(
            self.redis_live,
            lambda r: r.hset(
                redis_key,
                mapping={
                    "up": json.dumps([str(e.id) for e in endpoints_to_expand.keys()]),
                    "finish_time": datetime.now(tzutc()).isoformat(),
                },
            ),
        )

        async def _delete():
            async with self.db.begin_session() as db_sess:
                query = (
                    sa.update(EndpointRow)
                    .values({
                        "destroyed_at": sa.func.now(),
                        "lifecycle_stage": EndpointLifecycle.DESTROYED,
                    })
                    .where(EndpointRow.id.in_([e.id for e in endpoints_to_mark_terminated]))
                )
                await db_sess.execute(query)
                query = sa.delete(RoutingRow).where(
                    RoutingRow.session.in_(already_destroyed_sessions)
                )
                await db_sess.execute(query)

        await execute_with_retry(_delete)

        async with self.db.begin_readonly_session() as db_sess:
            for endpoint in endpoints_to_mark_terminated:
                try:
                    await self.registry.delete_appproxy_endpoint(
                        db_sess,
                        endpoint,
                    )
                except Exception as e:
                    log.warning("failed to communicate with AppProxy endpoint: {}", str(e))

    async def start_session(
        self,
        sched_ctx: SchedulingContext,
        session: SessionRow,
    ) -> None:
        log_fmt = (
            "prepare(s:{0.id}, type:{0.session_type}, name:{0.name}, "
            "ak:{0.access_key}, cluster_mode:{0.cluster_mode}): "
        )
        log_args = (session,)
        log.debug(log_fmt + "try-starting", *log_args)
        try:
            assert len(session.kernels) > 0
            await self.registry.start_session(sched_ctx, session)
        except Exception as e:
            status_data = convert_to_status_data(e, self.local_config["debug"]["enabled"])
            log.warning(log_fmt + "failed-starting", *log_args, exc_info=True)
            # TODO: instead of instantly cancelling upon exception, we could mark it as
            #       SCHEDULED and retry within some limit using status_data.

            async def _mark_session_cancelled() -> None:
                async with self.db.begin() as db_conn:
                    affected_agents = set(k.agent for k in session.kernels)
                    await _rollback_predicate_mutations(db_conn, sched_ctx, session)
                    now = datetime.now(tzutc())
                    update_query = (
                        sa.update(KernelRow)
                        .values(
                            status=KernelStatus.CANCELLED,
                            status_changed=now,
                            status_info="failed-to-start",
                            status_data=status_data,
                            terminated_at=now,
                            status_history=sql_json_merge(
                                KernelRow.status_history,
                                (),
                                {
                                    KernelStatus.CANCELLED.name: now.isoformat(),
                                },
                            ),
                        )
                        .where(KernelRow.session_id == session.id)
                    )
                    await SASession(db_conn).execute(update_query)
                    update_sess_query = (
                        sa.update(SessionRow)
                        .values(
                            status=SessionStatus.CANCELLED,
                            # status_changed=now,
                            status_info="failed-to-start",
                            status_data=status_data,
                            terminated_at=now,
                            status_history=sql_json_merge(
                                SessionRow.status_history,
                                (),
                                {
                                    SessionStatus.CANCELLED.name: now.isoformat(),
                                },
                            ),
                        )
                        .where(SessionRow.id == session.id)
                    )
                    await SASession(db_conn).execute(update_sess_query)
                    for agent_id in affected_agents:
                        await recalc_agent_resource_occupancy(db_conn, agent_id)

            log.debug(log_fmt + "cleanup-start-failure: begin", *log_args)
            try:
                await execute_with_retry(_mark_session_cancelled)
                await self.registry.event_producer.produce_event(
                    SessionCancelledEvent(
                        session.id,
                        session.creation_id,
                        KernelLifecycleEventReason.FAILED_TO_START,
                    ),
                )
                async with self.db.begin_readonly_session() as db_sess:
                    query = sa.select(KernelRow.id, KernelRow.container_id).where(
                        KernelRow.session_id == session.id
                    )
                    rows = (await db_sess.execute(query)).fetchall()
                    cid_map = {row["id"]: row["container_id"] for row in rows}
                destroyed_kernels = [
                    {
                        "agent": k.agent,
                        "agent_addr": k.agent_addr,
                        "id": k.id,
                        "container_id": cid_map[k.id],
                    }
                    for k in session.kernels
                ]
                await self.registry.destroy_session_lowlevel(
                    session.id,
                    destroyed_kernels,
                )
                await self.registry.recalc_resource_usage()
            except Exception as destroy_err:
                log.error(log_fmt + "cleanup-start-failure: error", *log_args, exc_info=destroy_err)
            finally:
                log.debug(log_fmt + "cleanup-start-failure: done", *log_args)
        else:
            log.info(log_fmt + "started", *log_args)


async def _list_managed_sessions(
    db_sess: SASession,
    sgroup_name: str,
    pending_timeout: timedelta,
) -> Tuple[List[SessionRow], List[SessionRow], List[SessionRow]]:
    """
    Return three lists of sessions.
    first is a list of existing sessions,
    second is pending sessions and third is to-be-cancelled sessions due to pending timeout.
    """

    managed_sessions = await SessionRow.get_sgroup_managed_sessions(db_sess, sgroup_name)

    candidates: List[SessionRow] = []
    cancelleds: List[SessionRow] = []
    existings: List[SessionRow] = []

    now = datetime.now(tzutc())
    key_func = lambda s: (s.status.value, s.created_at)
    for status, sessions in itertools.groupby(
        sorted(managed_sessions, key=key_func),
        key=lambda s: s.status,
    ):
        if status != SessionStatus.PENDING:
            existings.extend(sessions)
            continue
        for sess in sessions:
            elapsed_pending_time = now - sess.created_at
            if pending_timeout.total_seconds() > 0 and elapsed_pending_time >= pending_timeout:
                cancelleds.append(sess)
            else:
                candidates.append(sess)

    return existings, candidates, cancelleds


async def _reserve_agent(
    sched_ctx: SchedulingContext,
    db_sess: SASession,
    scaling_group: str,
    agent_id: Optional[AgentId],
    requested_slots: ResourceSlot,
    extra_conds: Any = None,
) -> AgentAllocationContext:
    query = sa.select(AgentRow.occupied_slots).where(AgentRow.id == agent_id).with_for_update()
    if extra_conds is not None:
        query = query.where(extra_conds)
    current_occupied_slots = (await db_sess.execute(query)).scalar()
    if current_occupied_slots is None:
        raise RuntimeError(f"No agent matching condition: {extra_conds}")
    update_query = (
        sa.update(AgentRow)
        .values(
            occupied_slots=(current_occupied_slots + requested_slots),
        )
        .where(AgentRow.id == agent_id)
    )
    await db_sess.execute(update_query)
    # Get the agent address for later RPC calls
    query = sa.select(AgentRow.addr).where(AgentRow.id == agent_id)
    agent_addr = await db_sess.scalar(query)
    assert agent_addr is not None
    return AgentAllocationContext(agent_id, agent_addr, scaling_group)


async def _rollback_predicate_mutations(
    db_sess: SASession,
    sched_ctx: SchedulingContext,
    session: SessionRow,
) -> None:
    """
    Rollback any changes performed by predicates.

    NOTE: We don't use the DB-level transaction rollback because we need to
    store the "ERROR" status to corresponding rows in the kernels table.
    """

    # Instead of decrementing concurrency_used, we recalculate the access_key's usage,
    # because asynchronous container launch failures and agent failures
    # (especially with multi-node multi-container cluster sessions)
    # may accumulate up multiple subtractions, resulting in
    # negative concurrency_occupied values.
    log.debug("recalculate concurrency used in rollback predicates (ak: {})", session.access_key)
    await recalc_concurrency_used(db_sess, sched_ctx.registry.redis_stat, session.access_key)



"""add image table

Revision ID: 015d84d5a5ef
Revises: 60a1effa77d2
Create Date: 2022-02-15 23:45:19.814677

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql import text

from ai.backend.manager.models.base import ForeignKeyIDColumn, IDColumn, convention
from ai.backend.manager.models.image import ImageType

# revision identifiers, used by Alembic.
revision = "015d84d5a5ef"
down_revision = "60a1effa77d2"
branch_labels = None
depends_on = None


def upgrade():
    metadata = sa.MetaData(naming_convention=convention)
    op.create_table(
        "images",
        metadata,
        IDColumn("id"),
        sa.Column("name", sa.String, index=True, nullable=False),
        sa.Column("image", sa.String, nullable=False, index=True),
        sa.Column(
            "created_at", sa.DateTime(timezone=True), server_default=sa.func.now(), index=True
        ),
        sa.Column("tag", sa.String, nullable=False, index=True),
        sa.Column("registry", sa.String, nullable=False, index=True),
        sa.Column("architecture", sa.String, nullable=False, server_default="x86_64", index=True),
        sa.Column("config_digest", sa.CHAR(length=72), nullable=False),
        sa.Column("size_bytes", sa.BigInteger, nullable=False),
        sa.Column("type", sa.Enum(ImageType, name="image_type"), nullable=False),
        sa.Column("accelerators", sa.String),
        sa.Column("labels", postgresql.JSONB(), nullable=False),
        sa.Column("resources", postgresql.JSONB(), nullable=False),
        sa.Index("ix_image_name_architecture", "name", "architecture", unique=True),
        sa.Index("ix_image_image_tag_registry", "image", "tag", "registry"),
    )

    op.create_table(
        "image_aliases",
        metadata,
        IDColumn("id"),
        sa.Column("alias", sa.String, unique=True, index=True),
        ForeignKeyIDColumn("image", "images.id", nullable=False),
        sa.Index("ix_image_alias_unique_ref", "image", "alias", unique=True),
    )


def downgrade():
    op.drop_table("image_aliases")
    op.drop_table("images")
    op.execute(text("DROP TYPE image_type"))



"""change char col to str

Revision ID: 11146ba02235
Revises: 0f7a4b643940
Create Date: 2022-03-25 12:32:05.637628

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "11146ba02235"
down_revision = "0f7a4b643940"
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()
    op.alter_column("agents", column_name="architecture", type_=sa.String(length=32))
    query = """
    UPDATE agents
    SET architecture = TRIM (architecture);
    """
    conn.execute(text(query))


def downgrade():
    op.alter_column("agents", column_name="architecture", type_=sa.CHAR(length=32))



"""add-kernels-uuid-prefix-index

Revision ID: f5530eccf202
Revises: ed666f476f39
Create Date: 2020-03-25 17:29:50.696450

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f5530eccf202"
down_revision = "ed666f476f39"
branch_labels = None
depends_on = None


def upgrade():
    op.create_index(
        op.f("ix_kernels_uuid_prefix"),
        "kernels",
        [sa.text('CAST("id" AS VARCHAR) COLLATE "C"')],
    )


def downgrade():
    op.drop_index(
        op.f("ix_kernels_uuid_prefix"),
        "kernels",
    )



"""allow_kernels_scaling_group_nullable

Revision ID: 9bd986a75a2a
Revises: 513164749de4
Create Date: 2019-09-20 14:39:57.761791

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9bd986a75a2a"
down_revision = "513164749de4"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "kernels",
        "scaling_group",
        existing_type=sa.VARCHAR(length=64),
        nullable=True,
        existing_server_default=sa.text("'default'::character varying"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "kernels",
        "scaling_group",
        existing_type=sa.VARCHAR(length=64),
        nullable=False,
        existing_server_default=sa.text("'default'::character varying"),
    )
    # ### end Alembic commands ###



"""add_images_col_to_session

Revision ID: eb9441fcf90a
Revises: 69c059996cbd
Create Date: 2023-07-06 13:51:52.098587

"""

from collections import defaultdict

import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm import registry
from sqlalchemy.sql.expression import bindparam

from ai.backend.manager.models.base import GUID, convention

# revision identifiers, used by Alembic.
revision = "eb9441fcf90a"
down_revision = "69c059996cbd"
branch_labels = None
depends_on = None

metadata = sa.MetaData(naming_convention=convention)
mapper_registry = registry(metadata=metadata)
Base = mapper_registry.generate_base()

PAGE_SIZE = 100
DEFAULT_ROLE = "main"


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    conn = op.get_bind()
    op.add_column("sessions", sa.Column("images", sa.ARRAY(sa.String()), nullable=True))
    kernels = sa.Table(
        "kernels",
        metadata,
        sa.Column("id", GUID, primary_key=True),
        sa.Column("session_id", GUID, nullable=False),
        sa.Column("image", sa.String(length=512)),
        sa.Column(
            "cluster_role",
            sa.String(length=16),
        ),
    )

    class SessionRow(Base):  # type: ignore[valid-type, misc]
        __tablename__ = "sessions"
        id = sa.Column(
            "id",
            GUID(),
            server_default=sa.text("uuid_generate_v4()"),
            nullable=False,
            primary_key=True,
        )
        images = sa.Column("images", sa.ARRAY(sa.String), nullable=True)

    kernel_cnt = conn.execute(sa.select([sa.func.count()]).select_from(kernels)).scalar()

    for offset in range(0, kernel_cnt, PAGE_SIZE):
        session_agent_ids_map = defaultdict(list)
        session_id_query = (
            sa.select([sa.distinct(kernels.c.session_id)])
            .where(kernels.c.image.is_not(None))
            .order_by(kernels.c.session_id)
            .offset(offset)
            .limit(PAGE_SIZE)
        )
        session_ids = conn.execute(session_id_query).scalars().all()
        stmt = sa.select([kernels]).where(kernels.c.session_id.in_(session_ids))
        kernel_rows = conn.execute(stmt).fetchall()
        for row in kernel_rows:
            img_list = session_agent_ids_map[row["session_id"]]
            if row["image"] in img_list:
                continue
            if row["cluster_role"] == DEFAULT_ROLE:
                img_list.insert(0, row["image"])
            else:
                img_list.append(row["image"])
        if session_agent_ids_map:
            conn.execute(
                sa.update(SessionRow)
                .values({"images": bindparam("images")})
                .where(SessionRow.id == bindparam("b_session_id")),
                [
                    {
                        "b_session_id": session_id,
                        "images": list(images),
                    }
                    for session_id, images in session_agent_ids_map.items()
                ],
            )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("sessions", "images")
    # ### end Alembic commands ###



from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Mapping, Optional, Sequence
from uuid import UUID

import attrs
import msgpack
import sqlalchemy as sa
from dateutil.tz.tz import tzfile
from redis.asyncio import Redis
from redis.asyncio.client import Pipeline as RedisPipeline
from sqlalchemy.orm import joinedload, load_only

from ai.backend.common import redis_helper
from ai.backend.common.types import RedisConnectionInfo
from ai.backend.common.utils import nmget

from .group import GroupRow
from .kernel import LIVE_STATUS, RESOURCE_USAGE_KERNEL_STATUSES, KernelRow, KernelStatus
from .session import SessionRow
from .user import UserRow
from .utils import ExtendedAsyncSAEngine

__all__: Sequence[str] = (
    "ResourceGroupUnit",
    "ResourceUsage",
    "BaseResourceUsageGroup",
    "parse_resource_usage",
    "parse_resource_usage_groups",
    "fetch_resource_usage",
)


class ResourceGroupUnit(str, Enum):
    KERNEL = "kernel"
    SESSION = "session"
    PROJECT = "project"
    DOMAIN = "domain"
    TOTAL = "total"


@attrs.define(slots=True)
class ResourceUsage:
    nfs: set = attrs.field(factory=set)
    cpu_allocated: float = attrs.field(default=0.0)
    cpu_used: float = attrs.field(default=0.0)
    mem_allocated: int = attrs.field(default=0)
    mem_used: int = attrs.field(default=0)
    shared_memory: int = attrs.field(default=0)
    disk_allocated: int = attrs.field(default=0)  # TODO: disk quota limit
    disk_used: int = attrs.field(default=0)
    io_read: int = attrs.field(default=0)
    io_write: int = attrs.field(default=0)
    device_type: set = attrs.field(factory=set)
    smp: float = attrs.field(default=0.0)
    gpu_mem_allocated: float = attrs.field(default=0.0)
    gpu_allocated: float = attrs.field(default=0.0)

    agent_ids: set[str] = attrs.field(factory=set)

    def __add__(self, other: Any) -> ResourceUsage:
        if not isinstance(other, ResourceUsage):
            raise TypeError(
                "ResourceUsage should only be added to `ResourceUsage` type,"
                f" not `{type(other)}` type."
            )
        assert isinstance(other, ResourceUsage)
        self.nfs |= other.nfs  # Is this correct??
        self.device_type |= other.device_type
        self.cpu_allocated += other.cpu_allocated
        self.cpu_used += other.cpu_used
        self.mem_allocated += other.mem_allocated
        self.mem_used += other.mem_used
        self.shared_memory += other.shared_memory
        self.disk_allocated += other.disk_allocated
        self.disk_used += other.disk_used
        self.io_read += other.io_read
        self.io_write += other.io_write
        self.smp += other.smp
        self.gpu_mem_allocated += other.gpu_mem_allocated
        self.gpu_allocated += other.gpu_allocated
        self.agent_ids |= other.agent_ids
        return self

    def to_json(self) -> Mapping[str, Any]:
        return {
            "nfs": list(self.nfs),
            "cpu_allocated": self.cpu_allocated,
            "cpu_used": self.cpu_used,
            "mem_allocated": self.mem_allocated,
            "mem_used": self.mem_used,
            "shared_memory": self.shared_memory,
            "disk_allocated": self.disk_allocated,
            "disk_used": self.disk_used,
            "io_read": self.io_read,
            "io_write": self.io_write,
            "device_type": list(self.device_type),
            "smp": self.smp,
            "gpu_mem_allocated": self.gpu_mem_allocated,
            "gpu_allocated": self.gpu_allocated,
        }

    def copy(self) -> ResourceUsage:
        return attrs.evolve(self, nfs={*self.nfs}, device_type={*self.device_type})


def to_str(val: Any) -> Optional[str]:
    return str(val) if val is not None else None


@attrs.define(slots=True)
class BaseResourceUsageGroup:
    group_unit: ResourceGroupUnit
    child_usage_group: dict = attrs.field(factory=dict)

    project_row: Optional[GroupRow] = attrs.field(default=None)
    session_row: Optional[SessionRow] = attrs.field(default=None)
    kernel_row: Optional[KernelRow] = attrs.field(default=None)

    created_at: Optional[datetime] = attrs.field(default=None)
    terminated_at: Optional[datetime] = attrs.field(default=None)
    scheduled_at: Optional[str] = attrs.field(default=None)
    used_time: Optional[str] = attrs.field(default=None)
    used_days: Optional[int] = attrs.field(default=None)
    agents: Optional[set[str]] = attrs.field(default=None)

    user_id: Optional[UUID] = attrs.field(default=None)
    user_email: Optional[str] = attrs.field(default=None)
    full_name: Optional[str] = attrs.field(default=None)  # User's full_name
    access_key: Optional[UUID] = attrs.field(default=None)
    project_id: Optional[UUID] = attrs.field(default=None)
    project_name: Optional[str] = attrs.field(default=None)
    kernel_id: Optional[UUID] = attrs.field(default=None)
    container_ids: Optional[set[str]] = attrs.field(default=None)
    session_id: Optional[UUID] = attrs.field(default=None)
    session_name: Optional[str] = attrs.field(default=None)
    domain_name: Optional[str] = attrs.field(default=None)
    images: Optional[set[str]] = attrs.field(default=None)

    last_stat: Optional[Mapping[str, Any]] = attrs.field(default=None)
    extra_info: Mapping[str, Any] = attrs.field(factory=dict)

    status: Optional[str] = attrs.field(default=None)
    status_info: Optional[str] = attrs.field(default=None)
    status_history: Optional[str] = attrs.field(default=None)
    cluster_mode: Optional[str] = attrs.field(default=None)

    total_usage: ResourceUsage = attrs.field(factory=ResourceUsage)

    # def to_json(self, child: bool = False) -> dict[str, Any]:
    #     return_val = {
    #         **self.to_map(),
    #         "agents": self.total_usage.agent_ids,
    #         "group_unit": self.group_unit.value,
    #         "total_usage": self.total_usage.to_json(),
    #     }
    #     if child:
    #         belonged_infos: dict[str, list[Mapping[str, Any]]] = {}
    #         for g in self.child_usage_group.values():
    #             if g.group_unit.value not in belonged_infos:
    #                 belonged_infos[g.group_unit.value] = []
    #             belonged_infos[g.group_unit.value].append(g.to_json())
    #         return_val = {
    #             **belonged_infos,
    #             **return_val,
    #             # **{k:v for k, v in belonged_infos.items()},
    #         }
    #     return return_val

    def to_map(self) -> dict[str, Any]:
        return {
            "created_at": self.created_at,
            "terminated_at": self.terminated_at,
            "used_time": self.used_time,
            "used_days": self.used_days,
            "user_id": self.user_id,
            "user_email": self.user_email,
            "access_key": self.access_key,
            "project_id": self.project_id,
            "project_name": self.project_name,
            "kernel_id": self.kernel_id,
            "container_ids": self.container_ids.copy() if self.container_ids is not None else None,
            "session_id": self.session_id,
            "session_name": self.session_name,
            "domain_name": self.domain_name,
            "last_stat": {**self.last_stat} if self.last_stat is not None else None,
            "extra_info": {**self.extra_info} if self.extra_info is not None else None,
            "full_name": self.full_name,
            "images": self.images.copy() if self.images is not None else None,
            "agents": self.agents.copy() if self.agents is not None else None,
            "status": self.status,
            "status_info": self.status_info,
            "status_history": self.status_history,
            "cluster_mode": self.cluster_mode,
            "scheduled_at": self.scheduled_at,
            "total_usage": self.total_usage.copy(),
        }

    def to_json_base(self) -> dict[str, Any]:
        return {
            "created_at": to_str(self.created_at),
            "used_time": self.used_time,
            "used_days": self.used_days,
            "user_id": to_str(self.user_id),
            "user_email": self.user_email,
            "access_key": self.access_key,
            "project_id": to_str(self.project_id),
            "project_name": self.project_name,
            "kernel_id": to_str(self.kernel_id),
            "container_ids": list(self.container_ids) if self.container_ids is not None else [],
            "session_id": to_str(self.session_id),
            "session_name": self.session_name,
            "domain_name": self.domain_name,
            "last_stat": self.last_stat,
            "extra_info": self.extra_info,
            "full_name": to_str(self.full_name),
            "images": list(self.images) if self.images is not None else [],
            "agents": list(self.agents) if self.agents is not None else [],
            "status": to_str(self.status),
            "status_info": to_str(self.status_info),
            "status_history": to_str(self.status_history),
            "cluster_mode": to_str(self.cluster_mode),
            "scheduled_at": to_str(self.scheduled_at),
            "terminated_at": to_str(self.terminated_at),
            "total_usage": self.total_usage.to_json(),
        }


@attrs.define(slots=True, kw_only=True)
class KernelResourceUsage(BaseResourceUsageGroup):
    group_unit: ResourceGroupUnit = ResourceGroupUnit.KERNEL
    # child_usage_group: dict = attrs.field(factory=dict)
    agent: str
    kernel_id: UUID
    project_row: GroupRow
    session_row: SessionRow
    kernel_row: KernelRow

    def to_json(self, child: bool = False) -> dict[str, Any]:
        return {
            **self.to_json_base(),
            "agents": list(self.agents) if self.agents is not None else [],
            "agent": self.agent,
            "group_unit": self.group_unit.value,
            "total_usage": self.total_usage.to_json(),
        }

    @classmethod
    def from_base_usage_group(cls, usage_group: BaseResourceUsageGroup) -> KernelResourceUsage:
        if usage_group.group_unit != ResourceGroupUnit.KERNEL:
            raise ValueError(
                "Unable to parse `KernelResourceUsage` from usage_group "
                "that DOES NOT have `ResourceGroupUnit.KERNEL` on group_unit field."
            )
        if (
            usage_group.project_row is None
            or usage_group.session_row is None
            or usage_group.kernel_row is None
        ):
            raise ValueError(
                "Cannot parse `KernelResourceUsage` from usage_group that have None value field"
            )
        return cls(
            project_row=usage_group.project_row,
            session_row=usage_group.session_row,
            kernel_row=usage_group.kernel_row,
            agent=usage_group.kernel_row.agent,
            **{
                **usage_group.to_map(),
                "total_usage": ResourceUsage(),
            },
        )

    def register_resource_group(
        self,
        other: BaseResourceUsageGroup,
    ) -> bool:
        return True


@attrs.define(slots=True, kw_only=True)
class SessionResourceUsage(BaseResourceUsageGroup):
    group_unit: ResourceGroupUnit = ResourceGroupUnit.SESSION
    child_usage_group: dict[UUID, KernelResourceUsage] = attrs.Factory(dict)
    session_id: UUID
    project_row: GroupRow
    session_row: SessionRow

    def to_json(self, child: bool = False) -> dict[str, Any]:
        return_val = {
            **self.to_json_base(),
            "agents": list(self.agents) if self.agents is not None else [],
            "group_unit": self.group_unit.value,
            "total_usage": self.total_usage.to_json(),
        }
        if child:
            return_val = {
                **return_val,
                "kernels": {
                    str(g.kernel_id): g.to_json(child) for g in self.child_usage_group.values()
                },
            }
        return return_val

    @classmethod
    def from_base_usage_group(cls, usage_group: BaseResourceUsageGroup) -> SessionResourceUsage:
        if usage_group.group_unit not in (ResourceGroupUnit.KERNEL, ResourceGroupUnit.SESSION):
            raise ValueError(
                "Unable to parse `SessionResourceUsage` from usage_group that DOES NOT have"
                " `ResourceGroupUnit.KERNEL` or `ResourceGroupUnit.SESSION` on group_unit field."
            )
        if usage_group.project_row is None or usage_group.session_row is None:
            raise ValueError(
                "Cannot parse `SessionResourceUsage` from usage_group that have None value field"
            )
        return cls(
            project_row=usage_group.project_row,
            session_row=usage_group.session_row,
            **{
                **usage_group.to_map(),
                "total_usage": ResourceUsage(),
            },
        )

    def register_resource_group(self, other: BaseResourceUsageGroup) -> bool:
        if other.kernel_id is None:
            return False
        if other.kernel_id in self.child_usage_group:
            return False
        self.child_usage_group[other.kernel_id] = KernelResourceUsage.from_base_usage_group(other)
        is_registered = self.child_usage_group[other.kernel_id].register_resource_group(other)
        if is_registered:
            self.total_usage += other.total_usage
            if self.images is None:
                self.images = {*(other.images or set())}
            else:
                self.images |= other.images or set()
            if self.agents is None:
                self.agents = {*(other.agents or set())}
            else:
                self.agents |= other.agents or set()
            if self.container_ids is None:
                self.container_ids = {*(other.container_ids or set())}
            else:
                self.container_ids |= other.container_ids or set()
        return is_registered


@attrs.define(slots=True, kw_only=True)
class ProjectResourceUsage(BaseResourceUsageGroup):
    group_unit: ResourceGroupUnit = ResourceGroupUnit.PROJECT
    child_usage_group: dict[UUID, SessionResourceUsage] = attrs.Factory(dict)
    project_id: UUID
    project_row: GroupRow

    def to_json(self, child: bool = False) -> dict[str, Any]:
        return_val = {
            **self.to_json_base(),
            "agents": list(self.total_usage.agent_ids),
            "group_unit": self.group_unit.value,
            "total_usage": self.total_usage.to_json(),
        }
        if child:
            return_val = {
                **return_val,
                "sessions": {
                    str(g.session_id): g.to_json(child) for g in self.child_usage_group.values()
                },
            }
        return return_val

    @classmethod
    def from_base_usage_group(cls, usage_group: BaseResourceUsageGroup) -> ProjectResourceUsage:
        if usage_group.project_row is None:
            raise ValueError(
                "Cannot parse `ProjectResourceUsage` from usage_group that have None value field"
            )
        filtered_data = {
            **usage_group.to_map(),
            "images": None,
            "full_name": None,
            "status": None,
            "status_info": None,
            "status_history": None,
            "cluster_mode": None,
            "scheduled_at": None,
            "terminated_at": None,
            "total_usage": ResourceUsage(),
        }
        return cls(
            project_row=usage_group.project_row,
            **filtered_data,
        )

    def register_resource_group(self, other: BaseResourceUsageGroup) -> bool:
        if other.session_id is None:
            return False
        if other.session_id not in self.child_usage_group:
            self.child_usage_group[other.session_id] = SessionResourceUsage.from_base_usage_group(
                other
            )
        is_registered = self.child_usage_group[other.session_id].register_resource_group(other)
        if is_registered:
            self.total_usage += other.total_usage
            if self.images is None:
                self.images = {*(other.images or set())}
            else:
                self.images |= other.images or set()
            if self.agents is None:
                self.agents = {*(other.agents or set())}
            else:
                self.agents |= other.agents or set()
            if self.container_ids is None:
                self.container_ids = {*(other.container_ids or set())}
            else:
                self.container_ids |= other.container_ids or set()
        return is_registered


# @attrs.define(slots=True, kw_only=True)
# class DomainResourceUsage(BaseResourceUsageGroup):
#     group_unit: ResourceGroupUnit = attrs.field(default=ResourceGroupUnit.DOMAIN)
#     child_usage_group: dict[UUID, ProjectResourceUsage] = attrs.Factory(dict)


def parse_total_resource_group(
    groups: Sequence[BaseResourceUsageGroup],
) -> tuple[dict[UUID, ProjectResourceUsage], ResourceUsage]:
    group_map: dict[UUID, ProjectResourceUsage] = {}
    total_usage = ResourceUsage()
    for g in groups:
        key = g.project_id
        if key is None:
            continue
        if key not in group_map:
            group_map[key] = ProjectResourceUsage.from_base_usage_group(g)
        is_registered = group_map[key].register_resource_group(g)
        if is_registered:
            total_usage += g.total_usage
    return group_map, total_usage


def parse_resource_usage(
    kernel: KernelRow,
    last_stat: Optional[Mapping[str, Any]],
) -> ResourceUsage:
    if not last_stat:
        return ResourceUsage(
            agent_ids={kernel.agent},
        )
    nfs = set()
    if kernel.vfolder_mounts:
        # For >=22.03, return used host directories instead of volume host, which is not so useful.
        nfs = set([str(mount.host_path) for mount in kernel.vfolder_mounts])
    elif kernel.mounts and isinstance(kernel.mounts[0], list):
        # For the kernel records that have legacy contents of `mounts`.
        nfs = set([mount[2] for mount in kernel.mounts])

    device_type = set()
    smp = 0
    gpu_mem_allocated = 0
    if kernel.attached_devices and "cuda" in kernel.attached_devices:
        for dev_info in kernel.attached_devices["cuda"]:
            if (model_name := dev_info.get("model_name")) is not None:
                device_type.add(model_name)
            smp += int(nmget(dev_info, "data.smp", 0))
            gpu_mem_allocated += int(nmget(dev_info, "data.mem", 0))
    gpu_allocated = 0
    if "cuda.devices" in kernel.occupied_slots:
        gpu_allocated = kernel.occupied_slots["cuda.devices"]
    if "cuda.shares" in kernel.occupied_slots:
        gpu_allocated = kernel.occupied_slots["cuda.shares"]

    return ResourceUsage(
        agent_ids={kernel.agent},
        nfs={*nfs},
        cpu_allocated=float(kernel.occupied_slots.get("cpu", 0)),
        cpu_used=float(nmget(last_stat, "cpu_used.current", 0)),
        mem_allocated=int(kernel.occupied_slots.get("mem", 0)),
        mem_used=int(nmget(last_stat, "mem.capacity", 0)),
        shared_memory=int(nmget(kernel.resource_opts, "shmem", 0)),
        disk_allocated=0,
        disk_used=int(nmget(last_stat, "io_scratch_size/stats.max", 0, "/")),
        io_read=int(nmget(last_stat, "io_read.current", 0)),
        io_write=int(nmget(last_stat, "io_write.current", 0)),
        device_type={*device_type},
        smp=float(smp),
        gpu_mem_allocated=float(gpu_mem_allocated),
        gpu_allocated=float(gpu_allocated),
    )


async def parse_resource_usage_groups(
    kernels: list[KernelRow],
    redis_stat: RedisConnectionInfo,
    local_tz: tzfile,
) -> list[BaseResourceUsageGroup]:
    stat_map = {k.id: k.last_stat for k in kernels}
    stat_empty_kerns = [k.id for k in kernels if not k.last_stat]

    async def _pipe_builder(r: Redis) -> RedisPipeline:
        pipe = r.pipeline()
        for kern_id in stat_empty_kerns:
            await pipe.get(str(kern_id))
        return pipe

    raw_stats = await redis_helper.execute(redis_stat, _pipe_builder)
    for kern_id, raw_stat in zip(stat_empty_kerns, raw_stats):
        if raw_stat is None:
            continue
        stat_map[kern_id] = msgpack.unpackb(raw_stat)

    return [
        BaseResourceUsageGroup(
            kernel_row=kern,
            project_row=kern.session.group,
            session_row=kern.session,
            created_at=kern.created_at,
            terminated_at=kern.terminated_at,
            scheduled_at=kern.status_history.get(KernelStatus.SCHEDULED.name),
            used_time=kern.used_time,
            used_days=kern.get_used_days(local_tz),
            last_stat=stat_map[kern.id],
            user_id=kern.session.user_uuid,
            user_email=kern.session.user.email,
            access_key=kern.session.access_key,
            project_id=kern.session.group.id,
            project_name=kern.session.group.name,
            kernel_id=kern.id,
            container_ids={kern.container_id},
            session_id=kern.session_id,
            session_name=kern.session.name,
            domain_name=kern.session.domain_name,
            full_name=kern.session.user.full_name,
            images={kern.image},
            agents={kern.agent},
            status=kern.status.name,
            status_history=kern.status_history,
            cluster_mode=kern.cluster_mode,
            status_info=kern.status_info,
            group_unit=ResourceGroupUnit.KERNEL,
            total_usage=parse_resource_usage(kern, stat_map[kern.id]),
        )
        for kern in kernels
    ]


SESSION_RESOURCE_SELECT_COLS = (
    SessionRow.created_at,
    SessionRow.user_uuid,
    SessionRow.name,
    SessionRow.domain_name,
    SessionRow.id,
    SessionRow.group_id,
    SessionRow.access_key,
    SessionRow.images,
    SessionRow.cluster_mode,
    SessionRow.status_history,
    SessionRow.status,
    SessionRow.status_info,
)

PROJECT_RESOURCE_SELECT_COLS = (
    GroupRow.created_at,
    GroupRow.id,
    GroupRow.name,
    GroupRow.domain_name,
)

KERNEL_RESOURCE_SELECT_COLS = (
    KernelRow.agent,
    KernelRow.created_at,
    KernelRow.terminated_at,
    KernelRow.last_stat,
    KernelRow.status_history,
    KernelRow.status,
    KernelRow.status_info,
    KernelRow.session_id,
    KernelRow.id,
    KernelRow.container_id,
    KernelRow.occupied_slots,
    KernelRow.attached_devices,
    KernelRow.vfolder_mounts,
    KernelRow.mounts,
    KernelRow.resource_opts,
    KernelRow.image,
    KernelRow.cluster_mode,
)


def _parse_query(
    kernel_cond: Optional[sa.sql.BinaryExpression] = None,
    session_cond: Optional[sa.sql.BinaryExpression] = None,
    project_cond: Optional[sa.sql.BinaryExpression] = None,
) -> sa.sql.Select:
    session_load = joinedload(KernelRow.session)
    if session_cond is not None:
        session_load = joinedload(KernelRow.session.and_(session_cond))

    project_load = joinedload(SessionRow.group)
    if project_cond is not None:
        project_load = joinedload(SessionRow.group.and_(project_cond))
    query = sa.select(KernelRow).options(
        load_only(*KERNEL_RESOURCE_SELECT_COLS),
        session_load.options(
            load_only(*SESSION_RESOURCE_SELECT_COLS),
            joinedload(SessionRow.user).options(
                load_only(UserRow.email, UserRow.username, UserRow.full_name)
            ),
            project_load.options(load_only(*PROJECT_RESOURCE_SELECT_COLS)),
        ),
    )
    if kernel_cond is not None:
        query = query.where(kernel_cond)
    return query


async def fetch_resource_usage(
    db_engine: ExtendedAsyncSAEngine,
    start_date: datetime,
    end_date: datetime,
    session_ids: Optional[Sequence[UUID]] = None,
    project_ids: Optional[Sequence[UUID]] = None,
) -> list[KernelRow]:
    project_cond = None
    if project_ids:
        project_cond = GroupRow.id.in_(project_ids)
    session_cond = None
    if session_ids:
        session_cond = SessionRow.id.in_(session_ids)
    query = _parse_query(
        kernel_cond=(
            # Filter sessions which existence period overlaps with requested period
            (
                (KernelRow.terminated_at >= start_date)
                & (KernelRow.created_at < end_date)
                & (KernelRow.status.in_(RESOURCE_USAGE_KERNEL_STATUSES))
            )
            |
            # Or, filter running sessions which created before requested end_date
            ((KernelRow.created_at < end_date) & (KernelRow.status.in_(LIVE_STATUS)))
        ),
        session_cond=session_cond,
        project_cond=project_cond,
    )
    async with db_engine.begin_readonly_session() as db_sess:
        result = await db_sess.execute(query)
        kernels = result.scalars().all()

    return kernels



from __future__ import annotations

import asyncio
import itertools
import logging
from contextlib import asynccontextmanager as actxmgr
from contextvars import ContextVar
from pathlib import PurePosixPath
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Final,
    Iterable,
    List,
    Mapping,
    Sequence,
    Tuple,
    TypedDict,
)

import aiohttp
import attrs
import graphene
import yarl

from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import HardwareMetadata, VFolderID

from ..api.exceptions import InvalidAPIParameters, VFolderOperationFailed
from ..exceptions import InvalidArgument
from .base import Item, PaginatedList

if TYPE_CHECKING:
    from .gql import GraphQueryContext

__all__ = (
    "StorageProxyInfo",
    "VolumeInfo",
    "StorageSessionManager",
    "StorageVolume",
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


@attrs.define(auto_attribs=True, slots=True, frozen=True)
class StorageProxyInfo:
    session: aiohttp.ClientSession
    secret: str
    client_api_url: yarl.URL
    manager_api_url: yarl.URL
    sftp_scaling_groups: list[str]


AUTH_TOKEN_HDR: Final = "X-BackendAI-Storage-Auth-Token"

_ctx_volumes_cache: ContextVar[List[Tuple[str, VolumeInfo]]] = ContextVar("_ctx_volumes")


class VolumeInfo(TypedDict):
    name: str
    backend: str
    path: str
    fsprefix: str
    capabilities: List[str]


class StorageSessionManager:
    _proxies: Mapping[str, StorageProxyInfo]
    _exposed_volume_info: List[str]

    def __init__(self, storage_config: Mapping[str, Any]) -> None:
        self.config = storage_config
        self._exposed_volume_info = self.config["exposed_volume_info"]
        self._proxies = {}
        for proxy_name, proxy_config in self.config["proxies"].items():
            connector = aiohttp.TCPConnector(ssl=proxy_config["ssl_verify"])
            session = aiohttp.ClientSession(connector=connector)
            self._proxies[proxy_name] = StorageProxyInfo(
                session=session,
                secret=proxy_config["secret"],
                client_api_url=yarl.URL(proxy_config["client_api"]),
                manager_api_url=yarl.URL(proxy_config["manager_api"]),
                sftp_scaling_groups=proxy_config["sftp_scaling_groups"],
            )

    async def aclose(self) -> None:
        close_aws = []
        for proxy_info in self._proxies.values():
            close_aws.append(proxy_info.session.close())
        await asyncio.gather(*close_aws, return_exceptions=True)

    @staticmethod
    def split_host(vfolder_host: str) -> Tuple[str, str]:
        proxy_name, _, volume_name = vfolder_host.partition(":")
        return proxy_name, volume_name

    async def get_all_volumes(self) -> Iterable[Tuple[str, VolumeInfo]]:
        """
        Returns a list of tuple
        [(proxy_name: str, volume_info: VolumeInfo), ...]
        """
        try:
            # per-asyncio-task cache
            return _ctx_volumes_cache.get()
        except LookupError:
            pass
        fetch_aws = []

        async def _fetch(
            proxy_name: str,
            proxy_info: StorageProxyInfo,
        ) -> Iterable[Tuple[str, VolumeInfo]]:
            async with proxy_info.session.request(
                "GET",
                proxy_info.manager_api_url / "volumes",
                raise_for_status=True,
                headers={AUTH_TOKEN_HDR: proxy_info.secret},
            ) as resp:
                reply = await resp.json()
                return ((proxy_name, volume_data) for volume_data in reply["volumes"])

        for proxy_name, proxy_info in self._proxies.items():
            fetch_aws.append(_fetch(proxy_name, proxy_info))
        results = [*itertools.chain(*await asyncio.gather(*fetch_aws))]
        _ctx_volumes_cache.set(results)
        return results

    async def get_sftp_scaling_groups(self, proxy_name: str) -> List[str]:
        if proxy_name not in self._proxies:
            raise IndexError(f"proxy {proxy_name} does not exist")
        return self._proxies[proxy_name].sftp_scaling_groups or []

    async def get_mount_path(
        self,
        vfolder_host: str,
        vfolder_id: VFolderID,
        subpath: PurePosixPath = PurePosixPath("."),
    ) -> str:
        async with self.request(
            vfolder_host,
            "GET",
            "folder/mount",
            json={
                "volume": self.split_host(vfolder_host)[1],
                "vfid": str(vfolder_id),
                "subpath": str(subpath),
            },
        ) as (_, resp):
            reply = await resp.json()
            return reply["path"]

    @actxmgr
    async def request(
        self,
        vfolder_host_or_proxy_name: str,
        method: str,
        request_relpath: str,
        *args,
        **kwargs,
    ) -> AsyncIterator[Tuple[yarl.URL, aiohttp.ClientResponse]]:
        proxy_name, _ = self.split_host(vfolder_host_or_proxy_name)
        try:
            proxy_info = self._proxies[proxy_name]
        except KeyError:
            raise InvalidArgument("There is no such storage proxy", proxy_name)
        headers = kwargs.pop("headers", {})
        headers[AUTH_TOKEN_HDR] = proxy_info.secret
        async with proxy_info.session.request(
            method,
            proxy_info.manager_api_url / request_relpath,
            *args,
            headers=headers,
            **kwargs,
        ) as client_resp:
            if client_resp.status // 100 != 2:
                try:
                    error_data = await client_resp.json()
                    raise VFolderOperationFailed(
                        extra_msg=error_data.pop("msg", None),
                        extra_data=error_data,
                    )
                except aiohttp.ClientResponseError:
                    # when the response body is not JSON, just raise with status info.
                    raise VFolderOperationFailed(
                        extra_msg=(
                            "Storage proxy responded with "
                            f"{client_resp.status} {client_resp.reason}"
                        ),
                        extra_data=None,
                    )
                except VFolderOperationFailed as e:
                    if client_resp.status // 100 == 5:
                        raise InvalidAPIParameters(e.extra_msg, e.extra_data)
                    # Raise as-is for semantic failures, not server errors.
                    raise
            yield proxy_info.client_api_url, client_resp


class StorageVolume(graphene.ObjectType):
    class Meta:
        interfaces = (Item,)

    # id: {proxy_name}:{name}
    backend = graphene.String()
    fsprefix = graphene.String()
    path = graphene.String()
    capabilities = graphene.List(graphene.String)
    hardware_metadata = graphene.JSONString()
    performance_metric = graphene.JSONString()
    usage = graphene.JSONString()
    proxy = graphene.String(
        description="Added since 24.03.0. Name of the proxy which this volume belongs to."
    )
    name = graphene.String(description="Added since 24.03.0. Name of the storage.")

    async def resolve_hardware_metadata(self, info: graphene.ResolveInfo) -> HardwareMetadata:
        ctx: GraphQueryContext = info.context
        return await ctx.registry.gather_storage_hwinfo(self.id)

    async def resolve_performance_metric(self, info: graphene.ResolveInfo) -> Mapping[str, Any]:
        ctx: GraphQueryContext = info.context
        proxy_name, volume_name = ctx.storage_manager.split_host(self.id)
        try:
            proxy_info = ctx.storage_manager._proxies[proxy_name]
        except KeyError:
            raise ValueError(f"no such storage proxy: {proxy_name!r}")
        try:
            async with proxy_info.session.request(
                "GET",
                proxy_info.manager_api_url / "volume/performance-metric",
                json={"volume": volume_name},
                raise_for_status=True,
                headers={AUTH_TOKEN_HDR: proxy_info.secret},
            ) as resp:
                reply = await resp.json()
                return reply["metric"]
        except aiohttp.ClientResponseError:
            return {}

    async def resolve_usage(self, info: graphene.ResolveInfo) -> Mapping[str, Any]:
        ctx: GraphQueryContext = info.context
        proxy_name, volume_name = ctx.storage_manager.split_host(self.id)
        try:
            proxy_info = ctx.storage_manager._proxies[proxy_name]
        except KeyError:
            raise ValueError(f"no such storage proxy: {proxy_name!r}")
        try:
            async with proxy_info.session.request(
                "GET",
                proxy_info.manager_api_url / "folder/fs-usage",
                json={"volume": volume_name},
                raise_for_status=True,
                headers={AUTH_TOKEN_HDR: proxy_info.secret},
            ) as resp:
                reply = await resp.json()
                return reply
        except aiohttp.ClientResponseError:
            return {}

    @classmethod
    def from_info(cls, proxy_name: str, volume_info: VolumeInfo) -> StorageVolume:
        return cls(
            id=f"{proxy_name}:{volume_info['name']}",
            backend=volume_info["backend"],
            path=volume_info["path"],
            fsprefix=volume_info["fsprefix"],
            capabilities=volume_info["capabilities"],
            name=volume_info["name"],
            proxy=proxy_name,
        )

    @classmethod
    async def load_count(
        cls,
        ctx: GraphQueryContext,
        filter: str = None,
    ) -> int:
        volumes = [*await ctx.storage_manager.get_all_volumes()]
        return len(volumes)

    @classmethod
    async def load_slice(
        cls,
        ctx: GraphQueryContext,
        limit: int,
        offset: int,
        filter: str = None,
        order: str = None,
    ) -> Sequence[StorageVolume]:
        # For consistency we add filter/order params here, but it's actually noop.
        if filter is not None or order is not None:
            log.warning(
                "Paginated list of storage volumes igonores custom filtering and/or ordering"
            )
        volumes = [*await ctx.storage_manager.get_all_volumes()]
        return [
            cls.from_info(proxy_name, volume_info)
            for proxy_name, volume_info in volumes[offset : offset + limit]
        ]

    @classmethod
    async def load_by_id(
        cls,
        ctx: GraphQueryContext,
        id: str,
    ) -> StorageVolume:
        proxy_name, volume_name = ctx.storage_manager.split_host(id)
        try:
            proxy_info = ctx.storage_manager._proxies[proxy_name]
        except KeyError:
            raise ValueError(f"no such storage proxy: {proxy_name!r}")
        async with proxy_info.session.request(
            "GET",
            proxy_info.manager_api_url / "volumes",
            raise_for_status=True,
            headers={AUTH_TOKEN_HDR: proxy_info.secret},
        ) as resp:
            reply = await resp.json()
            for volume_data in reply["volumes"]:
                if volume_data["name"] == volume_name:
                    return cls.from_info(proxy_name, volume_data)
            else:
                raise ValueError(
                    f"no such volume in the storage proxy {proxy_name!r}: {volume_name!r}",
                )


class StorageVolumeList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(StorageVolume, required=True)



"""empty message

Revision ID: 36b69ddee76e
# Revises: 3efd66393bd0, f108628f032b
Revises: f108628f032b
Create Date: 2023-02-22 06:46:02.379245

"""

# revision identifiers, used by Alembic.
revision = "36b69ddee76e"
down_revision = "f108628f032b"
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



"""add-quota_scope_id

Revision ID: 5fbd368d12a2
Revises: 85984c98b90f
Create Date: 2023-03-29 07:29:27.592211

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "5fbd368d12a2"
down_revision = "85984c98b90f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("vfolders", sa.Column("quota_scope_id", sa.String(length=64), nullable=True))
    op.execute(
        text(
            "UPDATE vfolders "
            "SET quota_scope_id = REPLACE(\"group\"::text, '-', '') "
            "WHERE ownership_type = 'group';"
        )
    )
    op.execute(
        text(
            "UPDATE vfolders "
            "SET quota_scope_id = REPLACE(\"user\"::text, '-', '') "
            "WHERE ownership_type = 'user';"
        )
    )
    op.alter_column("vfolders", "quota_scope_id", nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("vfolders", "quota_scope_id")
    # ### end Alembic commands ###



import datetime
import logging
import uuid
from enum import Enum
from typing import TYPE_CHECKING, Any, Iterable, List, Mapping, Optional, Sequence

import graphene
import jwt
import sqlalchemy as sa
import yarl
from graphene.types.datetime import DateTime as GQLDateTime
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import relationship, selectinload
from sqlalchemy.orm.exc import NoResultFound

from ai.backend.common.docker import ImageRef
from ai.backend.common.logging_utils import BraceStyleAdapter
from ai.backend.common.types import ClusterMode, ResourceSlot
from ai.backend.manager.defs import SERVICE_MAX_RETRIES

from ..api.exceptions import EndpointNotFound, EndpointTokenNotFound
from .base import (
    GUID,
    Base,
    EndpointIDColumn,
    EnumValueType,
    ForeignKeyIDColumn,
    IDColumn,
    InferenceSessionError,
    Item,
    PaginatedList,
    ResourceSlotColumn,
    URLColumn,
    set_if_set,
    simple_db_mutate_returning_item,
)
from .image import ImageNode, ImageRefType, ImageRow
from .routing import RouteStatus, Routing
from .user import UserRole

if TYPE_CHECKING:
    from .gql import GraphQueryContext

__all__ = (
    "EndpointRow",
    "Endpoint",
    "EndpointLifecycle",
    "EndpointList",
    "ModifyEndpoint",
    "EndpointTokenRow",
    "EndpointToken",
    "EndpointTokenList",
)


log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore


class EndpointLifecycle(Enum):
    CREATED = "created"
    DESTROYING = "destroying"
    DESTROYED = "destroyed"


class EndpointRow(Base):
    __tablename__ = "endpoints"

    id = EndpointIDColumn()
    name = sa.Column("name", sa.String(length=512), nullable=False)
    created_user = sa.Column(
        "created_user", GUID, sa.ForeignKey("users.uuid", ondelete="RESTRICT"), nullable=False
    )
    session_owner = sa.Column(
        "session_owner", GUID, sa.ForeignKey("users.uuid", ondelete="RESTRICT"), nullable=False
    )
    # minus session count means this endpoint is requested for removal
    desired_session_count = sa.Column(
        "desired_session_count", sa.Integer, nullable=False, default=0, server_default="0"
    )
    image = sa.Column(
        "image", GUID, sa.ForeignKey("images.id", ondelete="RESTRICT"), nullable=False
    )
    model = sa.Column(
        "model",
        GUID,
        sa.ForeignKey("vfolders.id", ondelete="SET NULL"),
        nullable=True,
    )
    model_mount_destiation = sa.Column(
        "model_mount_destiation",
        sa.String(length=1024),
        nullable=False,
        default="/models",
        server_default="/models",
    )
    domain = sa.Column(
        "domain",
        sa.String(length=64),
        sa.ForeignKey("domains.name", ondelete="RESTRICT"),
        nullable=False,
    )
    project = sa.Column(
        "project",
        GUID,
        sa.ForeignKey("groups.id", ondelete="RESTRICT"),
        nullable=False,
    )
    resource_group = sa.Column(
        "resource_group",
        sa.ForeignKey("scaling_groups.name", ondelete="RESTRICT"),
        index=True,
        nullable=False,
    )
    lifecycle_stage = sa.Column(
        "lifecycle_stage",
        EnumValueType(EndpointLifecycle),
        nullable=False,
        default=EndpointLifecycle.CREATED,
    )
    tag = sa.Column("tag", sa.String(length=64), nullable=True)
    startup_command = sa.Column("startup_command", sa.Text, nullable=True)
    bootstrap_script = sa.Column("bootstrap_script", sa.String(length=16 * 1024), nullable=True)
    callback_url = sa.Column("callback_url", URLColumn, nullable=True, default=sa.null())
    environ = sa.Column("environ", pgsql.JSONB(), nullable=True, default={})
    open_to_public = sa.Column("open_to_public", sa.Boolean, default=False)

    resource_slots = sa.Column("resource_slots", ResourceSlotColumn(), nullable=False)
    url = sa.Column("url", sa.String(length=1024))
    resource_opts = sa.Column("resource_opts", pgsql.JSONB(), nullable=True, default={})
    cluster_mode = sa.Column(
        "cluster_mode",
        sa.String(length=16),
        nullable=False,
        default=ClusterMode.SINGLE_NODE,
        server_default=ClusterMode.SINGLE_NODE.name,
    )
    cluster_size = sa.Column(
        "cluster_size", sa.Integer, nullable=False, default=1, server_default="1"
    )

    retries = sa.Column("retries", sa.Integer, nullable=False, default=0, server_default="0")
    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.text("now()"),
        nullable=True,
    )
    destroyed_at = sa.Column(
        "destroyed_at",
        sa.DateTime(timezone=True),
        nullable=True,
    )

    routings = relationship("RoutingRow", back_populates="endpoint_row")
    tokens = relationship("EndpointTokenRow", back_populates="endpoint_row")
    image_row = relationship("ImageRow", back_populates="endpoints")
    created_user_row = relationship(
        "UserRow", back_populates="created_endpoints", foreign_keys="EndpointRow.created_user"
    )
    session_owner_row = relationship(
        "UserRow", back_populates="owned_endpoints", foreign_keys="EndpointRow.session_owner"
    )

    def __init__(
        self,
        name: str,
        created_user: uuid.UUID,
        session_owner: uuid.UUID,
        desired_session_count: int,
        image: ImageRow,
        model: uuid.UUID,
        domain: str,
        project: uuid.UUID,
        resource_group: str,
        resource_slots: Mapping[str, Any],
        cluster_mode: ClusterMode,
        cluster_size: int,
        model_mount_destination: Optional[str] = None,
        tag: Optional[str] = None,
        startup_command: Optional[str] = None,
        bootstrap_script: Optional[str] = None,
        callback_url: Optional[yarl.URL] = None,
        environ: Optional[Mapping[str, Any]] = None,
        resource_opts: Optional[Mapping[str, Any]] = None,
        open_to_public=False,
    ):
        self.id = uuid.uuid4()
        self.name = name
        self.created_user = created_user
        self.session_owner = session_owner
        self.desired_session_count = desired_session_count
        self.image = image.id
        self.model = model
        self.domain = domain
        self.project = project
        self.resource_group = resource_group
        self.resource_slots = resource_slots
        self.cluster_mode = cluster_mode.name
        self.cluster_size = cluster_size
        self.model_mount_destination = model_mount_destination
        self.tag = tag
        self.startup_command = startup_command
        self.bootstrap_script = bootstrap_script
        self.callback_url = callback_url
        self.environ = environ
        self.resource_opts = resource_opts
        self.open_to_public = open_to_public

    @classmethod
    async def get(
        cls,
        session: AsyncSession,
        endpoint_id: uuid.UUID,
        domain: Optional[str] = None,
        project: Optional[uuid.UUID] = None,
        user_uuid: Optional[uuid.UUID] = None,
        load_routes=False,
        load_tokens=False,
        load_image=False,
        load_created_user=False,
        load_session_owner=False,
    ) -> "EndpointRow":
        """
        :raises: sqlalchemy.orm.exc.NoResultFound
        """
        query = sa.select(EndpointRow).filter(EndpointRow.id == endpoint_id)
        if load_routes:
            query = query.options(selectinload(EndpointRow.routings))
        if load_tokens:
            query = query.options(selectinload(EndpointRow.tokens))
        if load_image:
            query = query.options(selectinload(EndpointRow.image_row))
        if load_created_user:
            query = query.options(selectinload(EndpointRow.created_user_row))
        if load_session_owner:
            query = query.options(selectinload(EndpointRow.session_owner_row))
        if project:
            query = query.filter(EndpointRow.project == project)
        if domain:
            query = query.filter(EndpointRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointRow.session_owner == user_uuid)
        result = await session.execute(query)
        row = result.scalar()
        if row is None:
            raise NoResultFound
        return row

    @classmethod
    async def list(
        cls,
        session: AsyncSession,
        domain: Optional[str] = None,
        project: Optional[uuid.UUID] = None,
        user_uuid: Optional[uuid.UUID] = None,
        load_routes=False,
        load_image=False,
        load_tokens=False,
        load_created_user=False,
        load_session_owner=False,
        status_filter=[EndpointLifecycle.CREATED],
    ) -> List["EndpointRow"]:
        query = (
            sa.select(EndpointRow)
            .order_by(sa.desc(EndpointRow.created_at))
            .filter(EndpointRow.lifecycle_stage.in_(status_filter))
        )
        if load_routes:
            query = query.options(selectinload(EndpointRow.routings))
        if load_tokens:
            query = query.options(selectinload(EndpointRow.tokens))
        if load_image:
            query = query.options(selectinload(EndpointRow.image_row))
        if load_created_user:
            query = query.options(selectinload(EndpointRow.created_user_row))
        if load_session_owner:
            query = query.options(selectinload(EndpointRow.session_owner_row))
        if project:
            query = query.filter(EndpointRow.project == project)
        if domain:
            query = query.filter(EndpointRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointRow.session_owner == user_uuid)
        result = await session.execute(query)
        return result.scalars().all()

    @classmethod
    async def list_by_model(
        cls,
        session: AsyncSession,
        model_id: uuid.UUID,
        domain: Optional[str] = None,
        project: Optional[uuid.UUID] = None,
        user_uuid: Optional[uuid.UUID] = None,
        load_routes=False,
        load_image=False,
        load_tokens=False,
        load_created_user=False,
        load_session_owner=False,
        status_filter=[EndpointLifecycle.CREATED],
    ) -> List["EndpointRow"]:
        query = (
            sa.select(EndpointRow)
            .order_by(sa.desc(EndpointRow.created_at))
            .filter(
                EndpointRow.lifecycle_stage.in_(status_filter) & (EndpointRow.model == model_id)
            )
        )
        if load_routes:
            query = query.options(selectinload(EndpointRow.routings))
        if load_tokens:
            query = query.options(selectinload(EndpointRow.tokens))
        if load_image:
            query = query.options(selectinload(EndpointRow.image_row))
        if load_created_user:
            query = query.options(selectinload(EndpointRow.created_user_row))
        if load_session_owner:
            query = query.options(selectinload(EndpointRow.session_owner_row))
        if project:
            query = query.filter(EndpointRow.project == project)
        if domain:
            query = query.filter(EndpointRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointRow.session_owner == user_uuid)
        result = await session.execute(query)
        return result.scalars().all()


class EndpointTokenRow(Base):
    __tablename__ = "endpoint_tokens"

    id = IDColumn()
    token = sa.Column("token", sa.String(), nullable=False)
    endpoint = sa.Column(
        "endpoint", GUID, sa.ForeignKey("endpoints.id", ondelete="SET NULL"), nullable=True
    )
    session_owner = ForeignKeyIDColumn("session_owner", "users.uuid")
    domain = sa.Column(
        "domain",
        sa.String(length=64),
        sa.ForeignKey("domains.name", ondelete="CASCADE"),
        nullable=False,
    )
    project = sa.Column(
        "project",
        GUID,
        sa.ForeignKey("groups.id", ondelete="CASCADE"),
        nullable=False,
    )
    created_at = sa.Column(
        "created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
    )

    endpoint_row = relationship("EndpointRow", back_populates="tokens")

    def __init__(
        self,
        id: uuid.UUID,
        token: str,
        endpoint: uuid.UUID,
        domain: str,
        project: uuid.UUID,
        session_owner: uuid.UUID,
    ) -> None:
        self.id = id
        self.token = token
        self.endpoint = endpoint
        self.domain = domain
        self.project = project
        self.session_owner = session_owner

    @classmethod
    async def list(
        cls,
        session: AsyncSession,
        endpoint_id: uuid.UUID,
        *,
        domain: Optional[str] = None,
        project: Optional[uuid.UUID] = None,
        user_uuid: Optional[uuid.UUID] = None,
        load_endpoint=False,
    ) -> Iterable["EndpointTokenRow"]:
        query = (
            sa.select(EndpointTokenRow)
            .filter(EndpointTokenRow.endpoint == endpoint_id)
            .order_by(sa.desc(EndpointTokenRow.created_at))
        )
        if load_endpoint:
            query = query.options(selectinload(EndpointTokenRow.tokens))
        if project:
            query = query.filter(EndpointTokenRow.project == project)
        if domain:
            query = query.filter(EndpointTokenRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointTokenRow.session_owner == user_uuid)
        result = await session.execute(query)
        return result.scalars().all()

    @classmethod
    async def get(
        cls,
        session: AsyncSession,
        token: str,
        *,
        domain: Optional[str] = None,
        project: Optional[uuid.UUID] = None,
        user_uuid: Optional[uuid.UUID] = None,
        load_endpoint=False,
    ) -> "EndpointTokenRow":
        query = sa.select(EndpointTokenRow).filter(EndpointTokenRow.token == token)
        if load_endpoint:
            query = query.options(selectinload(EndpointTokenRow.tokens))
        if project:
            query = query.filter(EndpointTokenRow.project == project)
        if domain:
            query = query.filter(EndpointTokenRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointTokenRow.session_owner == user_uuid)
        result = await session.execute(query)
        row = result.scalar()
        if not row:
            raise NoResultFound
        return row


class Endpoint(graphene.ObjectType):
    class Meta:
        interfaces = (Item,)

    endpoint_id = graphene.UUID()
    image = graphene.String(deprecation_reason="Deprecated since 23.09.9; use `image_object`")
    image_object = graphene.Field(ImageNode, description="Added at 23.09.9")
    domain = graphene.String()
    project = graphene.String()
    resource_group = graphene.String()
    resource_slots = graphene.JSONString()
    url = graphene.String()
    model = graphene.UUID()
    model_mount_destiation = graphene.String()
    created_user = graphene.UUID(
        deprecation_reason="Deprecated since 23.09.8; use `created_user_id`"
    )
    created_user_email = graphene.String(description="Added at 23.09.8")
    created_user_id = graphene.UUID(description="Added at 23.09.8")
    session_owner = graphene.UUID(
        deprecation_reason="Deprecated since 23.09.8; use `session_owner_id`"
    )
    session_owner_email = graphene.String(description="Added at 23.09.8")
    session_owner_id = graphene.UUID(description="Added at 23.09.8")
    tag = graphene.String()
    startup_command = graphene.String()
    bootstrap_script = graphene.String()
    callback_url = graphene.String()
    environ = graphene.JSONString()
    name = graphene.String()
    resource_opts = graphene.JSONString()
    desired_session_count = graphene.Int()
    cluster_mode = graphene.String()
    cluster_size = graphene.Int()
    open_to_public = graphene.Boolean()

    created_at = GQLDateTime(required=True)
    destroyed_at = GQLDateTime()

    routings = graphene.List(Routing)
    retries = graphene.Int()
    status = graphene.String()

    lifecycle_stage = graphene.String()

    errors = graphene.List(graphene.NonNull(InferenceSessionError), required=True)

    @classmethod
    async def from_row(
        cls,
        ctx,  # ctx: GraphQueryContext,
        row: EndpointRow,
    ) -> "Endpoint":
        return cls(
            endpoint_id=row.id,
            # image="", # deprecated, row.image_object.name,
            image_object=ImageNode.from_row(row.image_row),
            domain=row.domain,
            project=row.project,
            resource_group=row.resource_group,
            resource_slots=row.resource_slots.to_json(),
            url=row.url,
            model=row.model,
            model_mount_destiation=row.model_mount_destiation,
            created_user=row.created_user,
            created_user_id=row.created_user,
            created_user_email=row.created_user_row.email,
            session_owner=row.session_owner,
            session_owner_id=row.session_owner,
            session_owner_email=row.session_owner_row.email,
            tag=row.tag,
            startup_command=row.startup_command,
            bootstrap_script=row.bootstrap_script,
            callback_url=row.callback_url,
            environ=row.environ,
            name=row.name,
            resource_opts=row.resource_opts,
            desired_session_count=row.desired_session_count,
            cluster_mode=row.cluster_mode,
            cluster_size=row.cluster_size,
            open_to_public=row.open_to_public,
            created_at=row.created_at,
            destroyed_at=row.destroyed_at,
            retries=row.retries,
            routings=[await Routing.from_row(None, r, endpoint=row) for r in row.routings],
            lifecycle_stage=row.lifecycle_stage.name,
        )

    @classmethod
    async def load_count(
        cls,
        ctx,  # ctx: GraphQueryContext,
        *,
        project: uuid.UUID | None = None,
        domain_name: Optional[str] = None,
        user_uuid: Optional[uuid.UUID] = None,
    ) -> int:
        query = (
            sa.select([sa.func.count()])
            .select_from(EndpointRow)
            .filter(
                EndpointRow.lifecycle_stage.in_([
                    EndpointLifecycle.CREATED,
                    EndpointLifecycle.DESTROYING,
                ])
            )
        )
        if project is not None:
            query = query.where(EndpointRow.project == project)
        if domain_name is not None:
            query = query.where(EndpointRow.domain == domain_name)
        if user_uuid is not None:
            query = query.where(EndpointRow.session_owner == user_uuid)
        async with ctx.db.begin_readonly() as conn:
            result = await conn.execute(query)
            return result.scalar()

    @classmethod
    async def load_slice(
        cls,
        ctx,  # ctx: GraphQueryContext,
        limit: int,
        offset: int,
        *,
        domain_name: Optional[str] = None,
        user_uuid: Optional[uuid.UUID] = None,
        project: Optional[uuid.UUID] = None,
        filter: Optional[str] = None,
        order: Optional[str] = None,
    ) -> Sequence["Endpoint"]:
        query = (
            sa.select(EndpointRow)
            .limit(limit)
            .offset(offset)
            .options(selectinload(EndpointRow.image_row))
            .options(selectinload(EndpointRow.routings))
            .options(selectinload(EndpointRow.created_user_row))
            .options(selectinload(EndpointRow.session_owner_row))
            .order_by(sa.desc(EndpointRow.created_at))
            .filter(
                EndpointRow.lifecycle_stage.in_([
                    EndpointLifecycle.CREATED,
                    EndpointLifecycle.DESTROYING,
                ])
            )
        )
        if project is not None:
            query = query.where(EndpointRow.project == project)
        if domain_name is not None:
            query = query.where(EndpointRow.domain == domain_name)
        if user_uuid is not None:
            query = query.where(EndpointRow.session_owner == user_uuid)
        """
        if filter is not None:
            parser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = parser.append_filter(query, filter)
        if order is not None:
            parser = QueryOrderParser(cls._queryorder_colmap)
            query = parser.append_ordering(query, order)
        """
        async with ctx.db.begin_readonly_session() as session:
            result = await session.execute(query)
            return [await cls.from_row(ctx, row) for row in result.scalars().all()]

    @classmethod
    async def load_all(
        cls,
        ctx,  # ctx: GraphQueryContext,
        *,
        domain_name: Optional[str] = None,
        user_uuid: Optional[uuid.UUID] = None,
        project: Optional[uuid.UUID] = None,
    ) -> Sequence["Endpoint"]:
        async with ctx.db.begin_readonly_session() as session:
            rows = await EndpointRow.list(
                session,
                project=project,
                domain=domain_name,
                user_uuid=user_uuid,
                load_image=True,
                load_created_user=True,
                load_session_owner=True,
            )
            return [await Endpoint.from_row(ctx, row) for row in rows]

    @classmethod
    async def load_item(
        cls,
        ctx,  # ctx: GraphQueryContext,
        *,
        endpoint_id: uuid.UUID,
        domain_name: Optional[str] = None,
        user_uuid: Optional[uuid.UUID] = None,
        project: uuid.UUID | None = None,
    ) -> "Endpoint":
        """
        :raises: ai.backend.manager.api.exceptions.EndpointNotFound
        """
        try:
            async with ctx.db.begin_readonly_session() as session:
                row = await EndpointRow.get(
                    session,
                    endpoint_id=endpoint_id,
                    domain=domain_name,
                    user_uuid=user_uuid,
                    project=project,
                    load_image=True,
                    load_routes=True,
                    load_created_user=True,
                    load_session_owner=True,
                )
                return await Endpoint.from_row(ctx, row)
        except NoResultFound:
            raise EndpointNotFound

    async def resolve_status(self, info: graphene.ResolveInfo) -> str:
        if self.retries > SERVICE_MAX_RETRIES:
            return "UNHEALTHY"
        if self.lifecycle_stage == EndpointLifecycle.DESTROYING:
            return "DESTROYING"
        if len(self.routings) == 0:
            return "READY"
        if (spawned_service_count := len([r for r in self.routings])) > 0:
            healthy_service_count = len([
                r for r in self.routings if r.status == RouteStatus.HEALTHY.name
            ])
            if healthy_service_count == spawned_service_count:
                return "HEALTHY"
            unhealthy_service_count = len([
                r for r in self.routings if r.status == RouteStatus.UNHEALTHY.name
            ])
            if unhealthy_service_count > 0:
                return "DEGRADED"
        return "PROVISIONING"

    async def resolve_errors(self, info: graphene.ResolveInfo) -> Any:
        error_routes = [r for r in self.routings if r.status == RouteStatus.FAILED_TO_START.name]
        errors = []
        for route in error_routes:
            match route.error_data["type"]:
                case "session_cancelled":
                    session_id = route.error_data["session_id"]
                case _:
                    session_id = None
            errors.append(
                InferenceSessionError(
                    session_id=session_id,
                    errors=[
                        InferenceSessionError.InferenceSessionErrorInfo(
                            src=e["src"], name=e["name"], repr=e["repr"]
                        )
                        for e in route.error_data["errors"]
                    ],
                )
            )

        return errors


class EndpointList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(Endpoint, required=True)


class ModifyEndpointInput(graphene.InputObjectType):
    resource_slots = graphene.JSONString()
    resource_opts = graphene.JSONString()
    cluster_mode = graphene.String()
    cluster_size = graphene.Int()
    desired_session_count = graphene.Int()
    image = ImageRefType()
    name = graphene.String()
    resource_group = graphene.String()
    open_to_public = graphene.Boolean()


class ModifyEndpoint(graphene.Mutation):
    class Arguments:
        endpoint_id = graphene.UUID(required=True)
        props = ModifyEndpointInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    endpoint = graphene.Field(lambda: Endpoint, required=False, description="Added at 23.09.8")

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        endpoint_id: uuid.UUID,
        props: ModifyEndpointInput,
    ) -> "ModifyEndpoint":
        graph_ctx: GraphQueryContext = info.context
        data: dict[str, Any] = {}
        set_if_set(
            props,
            data,
            "resource_slots",
            clean_func=lambda v: ResourceSlot.from_user_input(v, None),
        )
        set_if_set(props, data, "resource_opts")
        set_if_set(props, data, "cluster_mode")
        set_if_set(props, data, "cluster_size")
        set_if_set(props, data, "desired_session_count")
        set_if_set(props, data, "image")
        set_if_set(props, data, "resource_group")
        image = data.pop("image", None)

        async with graph_ctx.db.begin_readonly_session() as db_session:
            match graph_ctx.user["role"]:
                case UserRole.SUPERADMIN:
                    pass
                case UserRole.ADMIN:
                    domain_name = graph_ctx.user["domain_name"]
                    stmt = (
                        sa.select(EndpointRow)
                        .where(EndpointRow.id == endpoint_id)
                        .where(EndpointRow.domain == domain_name)
                    )
                    endpoint_row = (await db_session.scalars(stmt)).first()
                    if endpoint_row is None:
                        raise EndpointNotFound
                case _:
                    user_id = graph_ctx.user["uuid"]
                    stmt = (
                        sa.select(EndpointRow)
                        .where(EndpointRow.id == endpoint_id)
                        .where(
                            (EndpointRow.session_owner == user_id)
                            | (EndpointRow.created_user == user_id)
                        )
                    )
                    endpoint_row = (await db_session.scalars(stmt)).first()
                    if endpoint_row is None:
                        raise EndpointNotFound

            if image is not None:
                image_name = image["name"]
                registry = image.get("registry") or ["*"]
                arch = image.get("architecture")
                if arch is not None:
                    image_ref = ImageRef(image_name, registry, arch)
                else:
                    image_ref = ImageRef(
                        image_name,
                        registry,
                    )
                image_object = await ImageRow.resolve(db_session, [image_ref])
                data["image"] = image_object.id

        update_query = sa.update(EndpointRow).values(**data).where(EndpointRow.id == endpoint_id)

        async def _post(conn, result) -> EndpointRow:
            endpoint = result.first()
            try:
                async with AsyncSession(conn) as session:
                    row = await EndpointRow.get(
                        session,
                        endpoint_id=endpoint.id,
                        domain=endpoint.domain,
                        user_uuid=endpoint.created_user,
                        project=endpoint.project,
                        load_image=True,
                        load_routes=True,
                        load_created_user=True,
                        load_session_owner=True,
                    )
                return row
            except NoResultFound:
                raise EndpointNotFound

        return await simple_db_mutate_returning_item(
            cls, graph_ctx, update_query, item_cls=Endpoint, post_func=_post
        )


class EndpointToken(graphene.ObjectType):
    class Meta:
        interfaces = (Item,)

    token = graphene.String(required=True)
    endpoint_id = graphene.UUID(required=True)
    domain = graphene.String(required=True)
    project = graphene.String(required=True)
    session_owner = graphene.UUID(required=True)

    created_at = GQLDateTime(required=True)
    valid_until = GQLDateTime()

    @classmethod
    async def from_row(
        cls,
        ctx,  # ctx: GraphQueryContext,
        row: EndpointTokenRow,
    ) -> "EndpointToken":
        return cls(
            token=row.token,
            endpoint_id=row.endpoint,
            domain=row.domain,
            project=row.project,
            session_owner=row.session_owner,
            created_at=row.created_at,
        )

    @classmethod
    async def load_count(
        cls,
        ctx,  # ctx: GraphQueryContext,
        *,
        endpoint_id: Optional[uuid.UUID] = None,
        project: Optional[uuid.UUID] = None,
        domain_name: Optional[str] = None,
        user_uuid: Optional[uuid.UUID] = None,
    ) -> int:
        query = sa.select([sa.func.count()]).select_from(EndpointTokenRow)
        if endpoint_id is not None:
            query = query.where(EndpointTokenRow.endpoint == endpoint_id)
        if project:
            query = query.filter(EndpointTokenRow.project == project)
        if domain_name:
            query = query.filter(EndpointTokenRow.domain == domain_name)
        if user_uuid:
            query = query.filter(EndpointTokenRow.session_owner == user_uuid)
        async with ctx.db.begin_readonly() as conn:
            result = await conn.execute(query)
            return result.scalar()

    @classmethod
    async def load_slice(
        cls,
        ctx,  # ctx: GraphQueryContext,
        limit: int,
        offset: int,
        *,
        endpoint_id: Optional[uuid.UUID] = None,
        filter: str | None = None,
        order: str | None = None,
        project: Optional[uuid.UUID] = None,
        domain_name: Optional[str] = None,
        user_uuid: Optional[uuid.UUID] = None,
    ) -> Sequence["EndpointToken"]:
        query = (
            sa.select(EndpointTokenRow)
            .limit(limit)
            .offset(offset)
            .order_by(sa.desc(EndpointTokenRow.created_at))
        )
        if endpoint_id is not None:
            query = query.where(EndpointTokenRow.endpoint == endpoint_id)
        if project:
            query = query.filter(EndpointTokenRow.project == project)
        if domain_name:
            query = query.filter(EndpointTokenRow.domain == domain_name)
        if user_uuid:
            query = query.filter(EndpointTokenRow.session_owner == user_uuid)
        """
        if filter is not None:
            parser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = parser.append_filter(query, filter)
        if order is not None:
            parser = QueryOrderParser(cls._queryorder_colmap)
            query = parser.append_ordering(query, order)
        """
        async with ctx.db.begin_readonly_session() as session:
            result = await session.execute(query)
            return [await cls.from_row(ctx, row) for row in result.scalars().all()]

    @classmethod
    async def load_all(
        cls,
        ctx,  # ctx: GraphQueryContext
        endpoint_id: uuid.UUID,
        *,
        project: Optional[uuid.UUID] = None,
        domain_name: Optional[str] = None,
        user_uuid: Optional[uuid.UUID] = None,
    ) -> Sequence["EndpointToken"]:
        async with ctx.db.begin_readonly_session() as session:
            rows = await EndpointTokenRow.list(
                session,
                endpoint_id,
                project=project,
                domain=domain_name,
                user_uuid=user_uuid,
            )
        return [await EndpointToken.from_row(ctx, row) for row in rows]

    @classmethod
    async def load_item(
        cls,
        ctx,  # ctx: GraphQueryContext,
        token: str,
        *,
        project: Optional[uuid.UUID] = None,
        domain_name: Optional[str] = None,
        user_uuid: Optional[uuid.UUID] = None,
    ) -> "EndpointToken":
        try:
            async with ctx.db.begin_readonly_session() as session:
                row = await EndpointTokenRow.get(
                    session, token, project=project, domain=domain_name, user_uuid=user_uuid
                )
        except NoResultFound:
            raise EndpointTokenNotFound
        return await EndpointToken.from_row(ctx, row)

    async def resolve_valid_until(
        self,
        info: graphene.ResolveInfo,
    ) -> datetime.datetime | None:
        try:
            decoded = jwt.decode(
                self.token,
                algorithms=["HS256"],
                options={"verify_signature": False, "verify_exp": False},
            )
        except jwt.DecodeError:
            return None
        if "exp" not in decoded:
            return None
        return datetime.datetime.fromtimestamp(decoded["exp"])


class EndpointTokenList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(EndpointToken, required=True)
